# future
from __future__ import annotations

# stdlib
from collections.abc import Sequence
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
import numpy as np
from numpy.typing import NDArray
from scipy.ndimage.interpolation import rotate

# relative
from .... import lib
from ....ast.klass import pointerize_args_and_kwargs
from ....core.adp.data_subject_ledger import DataSubjectLedger
from ....core.adp.data_subject_list import DataSubjectArray
from ....core.adp.data_subject_list import dslarraytonumpyutf8
from ....core.adp.data_subject_list import numpyutf8todslarray
from ....core.node.common.action.get_or_set_property_action import (
    GetOrSetPropertyAction,
)
from ....core.node.common.action.get_or_set_property_action import PropertyActions
from ....lib.numpy.array import capnp_deserialize
from ....lib.numpy.array import capnp_serialize
from ....lib.python.util import upcast
from ....util import inherit_tags
from ...common.serde.capnp import CapnpModule
from ...common.serde.capnp import chunk_bytes
from ...common.serde.capnp import combine_bytes
from ...common.serde.capnp import get_capnp_schema
from ...common.serde.capnp import serde_magic_header
from ...common.serde.deserialize import _deserialize as deserialize
from ...common.serde.serializable import serializable
from ...common.serde.serialize import _serialize as serialize
from ...common.uid import UID
from ...node.abstract.node import AbstractNodeClient
from ...node.common.action.run_class_method_action import RunClassMethodAction
from ...pointer.pointer import Pointer
from ..broadcastable import is_broadcastable
from ..config import DEFAULT_INT_NUMPY_TYPE
from ..fixed_precision_tensor import FixedPrecisionTensor
from ..lazy_repeat_array import compute_min_max
from ..lazy_repeat_array import lazyrepeatarray
from ..passthrough import AcceptableSimpleType  # type: ignore
from ..passthrough import PassthroughTensor  # type: ignore
from ..passthrough import SupportedChainType  # type: ignore
from ..passthrough import is_acceptable_simple_type  # type: ignore
from ..smpc import utils
from ..smpc.mpc_tensor import MPCTensor
from ..smpc.utils import TYPE_TO_RING_SIZE
from ..util import implements
from .adp_tensor import ADPTensor
from .gamma_tensor import GammaTensor
from .gamma_tensor import TensorWrappedGammaTensorPointer


@serializable(recursive_serde=True)
class TensorWrappedPhiTensorPointer(Pointer, PassthroughTensor):
    __name__ = "TensorWrappedPhiTensorPointer"
    __module__ = "syft.core.tensor.autodp.phi_tensor"
    __attr_allowlist__ = [
        # default pointer attrs
        "client",
        "id_at_location",
        "object_type",
        "tags",
        "description",
        # phi_tensor attrs
        "data_subjects",
        "min_vals",
        "max_vals",
        "public_dtype",
        "public_shape",
    ]

    __serde_overrides__ = {
        "client": [lambda x: x.address, lambda y: y],
        "public_shape": [lambda x: x, lambda y: upcast(y)],
        "data_subjects": [dslarraytonumpyutf8, numpyutf8todslarray],
    }
    _exhausted = False
    is_enum = False

    def __init__(
        self,
        data_subjects: np.ndarray,
        min_vals: np.typing.ArrayLike,
        max_vals: np.typing.ArrayLike,
        client: Any,
        id_at_location: Optional[UID] = None,
        object_type: str = "",
        tags: Optional[List[str]] = None,
        description: str = "",
        public_shape: Optional[Tuple[int, ...]] = None,
        public_dtype: Optional[np.dtype] = None,
    ):
        super().__init__(
            client=client,
            id_at_location=id_at_location,
            object_type=object_type,
            tags=tags,
            description=description,
        )

        self.min_vals = min_vals
        self.max_vals = max_vals
        self.data_subjects = data_subjects
        self.public_shape = public_shape
        self.public_dtype = public_dtype

    # TODO: Modify for large arrays
    @property
    def synthetic(self) -> np.ndarray:
        public_dtype_func = getattr(
            self.public_dtype, "upcast", lambda: self.public_dtype
        )
        return (
            np.random.rand(*list(self.public_shape))  # type: ignore
            * (self.max_vals.to_numpy() - self.min_vals.to_numpy())
            + self.min_vals.to_numpy()
        ).astype(public_dtype_func())

    def __repr__(self) -> str:
        return (
            self.synthetic.__repr__()
            + "\n\n (The data printed above is synthetic - it's an imitation of the real data.)"
        )

    def share(self, *parties: Tuple[AbstractNodeClient, ...]) -> MPCTensor:
        all_parties = list(parties) + [self.client]
        ring_size = TYPE_TO_RING_SIZE.get(self.public_dtype, None)
        self_mpc = MPCTensor(
            secret=self,
            shape=self.public_shape,
            ring_size=ring_size,
            parties=all_parties,
        )
        return self_mpc

    @property
    def shape(self) -> Optional[Tuple[int, ...]]:
        if hasattr(self, "public_shape"):
            return self.public_shape
        else:
            return None

    def _apply_tensor_op(self, other: Any, op_str: str) -> Any:
        # we want to get the return type which matches the attr_path_and_name
        # so we ask lib_ast for the return type name that matches out
        # attr_path_and_name and then use that to get the actual pointer klass
        # then set the result to that pointer klass

        # We always maintain a Tensor hierarchy Tensor ---> PT--> Actual Data
        attr_path_and_name = f"syft.core.tensor.tensor.Tensor.{op_str}"
        min_vals, max_vals = compute_min_max(
            self.min_vals, self.max_vals, other, op_str
        )
        result = TensorWrappedPhiTensorPointer(
            data_subjects=self.data_subjects,
            min_vals=min_vals,
            max_vals=max_vals,
            client=self.client,
        )

        # QUESTION can the id_at_location be None?
        result_id_at_location = getattr(result, "id_at_location", None)

        if result_id_at_location is not None:
            # first downcast anything primitive which is not already PyPrimitive
            (
                downcast_args,
                downcast_kwargs,
            ) = lib.python.util.downcast_args_and_kwargs(args=[other], kwargs={})

            # then we convert anything which isnt a pointer into a pointer
            pointer_args, pointer_kwargs = pointerize_args_and_kwargs(
                args=downcast_args,
                kwargs=downcast_kwargs,
                client=self.client,
                gc_enabled=False,
            )

            cmd = RunClassMethodAction(
                path=attr_path_and_name,
                _self=self,
                args=pointer_args,
                kwargs=pointer_kwargs,
                id_at_location=result_id_at_location,
                address=self.client.address,
            )
            self.client.send_immediate_msg_without_reply(msg=cmd)

        inherit_tags(
            attr_path_and_name=attr_path_and_name,
            result=result,
            self_obj=self,
            args=[other],
            kwargs={},
        )

        result_public_shape = None

        if isinstance(other, TensorWrappedPhiTensorPointer):
            other_shape = other.public_shape
            other_dtype = other.public_dtype
        elif isinstance(other, (int, float)):
            other_shape = (1,)
            other_dtype = DEFAULT_INT_NUMPY_TYPE
        elif isinstance(other, bool):
            other_shape = (1,)
            other_dtype = np.dtype("bool")
        elif isinstance(other, np.ndarray):
            other_shape = other.shape
            other_dtype = other.dtype
        else:
            raise ValueError(
                f"Invalid Type for TensorWrappedPhiTensorPointer:{type(other)}"
            )

        if self.public_shape is not None and other_shape is not None:
            result_public_shape = utils.get_shape(
                op_str, self.public_shape, other_shape
            )

        if self.public_dtype is None or other_dtype is None:
            if self.public_dtype != other_dtype:
                raise ValueError(
                    f"Dtype for self: {self.public_dtype} and other :{other_dtype} should not be None"
                )
        result_public_dtype = self.public_dtype

        result.public_shape = result_public_shape
        result.public_dtype = result_public_dtype

        return result

    @property
    def gamma(self) -> TensorWrappedGammaTensorPointer:
        return TensorWrappedGammaTensorPointer(
            data_subjects=self.data_subjects,
            client=self.client,
            id_at_location=self.id_at_location,
            object_type=self.object_type,
            tags=self.tags,
            description=self.description,
            min_vals=self.min_vals,
            max_vals=self.max_vals,
            public_shape=getattr(self, "public_shape", None),
            public_dtype=getattr(self, "public_dtype", None),
        )

    @staticmethod
    def _apply_op(
        self: TensorWrappedPhiTensorPointer,
        other: Union[TensorWrappedPhiTensorPointer, MPCTensor, int, float, np.ndarray],
        op_str: str,
    ) -> Union[MPCTensor, TensorWrappedPhiTensorPointer]:
        """Performs the operation based on op_str

        Args:
            other (Union[TensorWrappedPhiTensorPointer,MPCTensor,int,float,np.ndarray]): second operand.

        Returns:
            Tuple[MPCTensor,Union[MPCTensor,int,float,np.ndarray]] : Result of the operation
        """
        if isinstance(other, TensorWrappedPhiTensorPointer):
            if (self.data_subjects != other.data_subjects).all():  # type: ignore
                return getattr(self.gamma, op_str)(other.gamma)
        elif isinstance(other, TensorWrappedGammaTensorPointer):
            return getattr(self.gamma, op_str)(other)

        if (
            isinstance(other, TensorWrappedPhiTensorPointer)
            and self.client != other.client
        ):

            parties = [self.client, other.client]

            self_mpc = MPCTensor(secret=self, shape=self.public_shape, parties=parties)
            other_mpc = MPCTensor(
                secret=other, shape=other.public_shape, parties=parties
            )

            return getattr(self_mpc, op_str)(other_mpc)

        elif isinstance(other, MPCTensor):

            return getattr(other, op_str)(self)
        elif is_acceptable_simple_type(other) or isinstance(
            other, TensorWrappedPhiTensorPointer
        ):
            return self._apply_tensor_op(other=other, op_str=op_str)
        else:
            print("Type is unsupported:" + str(type(other)))
            raise NotImplementedError

    def __add__(
        self,
        other: Union[TensorWrappedPhiTensorPointer, MPCTensor, int, float, np.ndarray],
    ) -> Union[TensorWrappedPhiTensorPointer, MPCTensor]:
        """Apply the "add" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorWrappedPhiTensorPointer._apply_op(self, other, "__add__")

    def __sub__(
        self,
        other: Union[TensorWrappedPhiTensorPointer, MPCTensor, int, float, np.ndarray],
    ) -> Union[TensorWrappedPhiTensorPointer, MPCTensor]:
        """Apply the "sub" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorWrappedPhiTensorPointer._apply_op(self, other, "__sub__")

    def __mul__(
        self,
        other: Union[TensorWrappedPhiTensorPointer, MPCTensor, int, float, np.ndarray],
    ) -> Union[TensorWrappedPhiTensorPointer, MPCTensor]:
        """Apply the "mul" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorWrappedPhiTensorPointer._apply_op(self, other, "__mul__")

    def __matmul__(
        self,
        other: Union[TensorWrappedPhiTensorPointer, MPCTensor, int, float, np.ndarray],
    ) -> Union[TensorWrappedPhiTensorPointer, MPCTensor]:
        """Apply the "matmul" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorWrappedPhiTensorPointer._apply_op(self, other, "__matmul__")

    def __rmatmul__(
        self,
        other: Union[TensorWrappedPhiTensorPointer, MPCTensor, int, float, np.ndarray],
    ) -> Union[TensorWrappedPhiTensorPointer, MPCTensor]:
        """Apply the "matmul" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorWrappedPhiTensorPointer._apply_op(self, other, "__rmatmul__")

    def __lt__(
        self,
        other: Union[TensorWrappedPhiTensorPointer, MPCTensor, int, float, np.ndarray],
    ) -> Union[TensorWrappedPhiTensorPointer, MPCTensor]:
        """Apply the "lt" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorWrappedPhiTensorPointer._apply_op(self, other, "__lt__")

    def __gt__(
        self,
        other: Union[TensorWrappedPhiTensorPointer, MPCTensor, int, float, np.ndarray],
    ) -> Union[TensorWrappedPhiTensorPointer, MPCTensor]:
        """Apply the "gt" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorWrappedPhiTensorPointer._apply_op(self, other, "__gt__")

    def __ge__(
        self,
        other: Union[TensorWrappedPhiTensorPointer, MPCTensor, int, float, np.ndarray],
    ) -> Union[TensorWrappedPhiTensorPointer, MPCTensor]:
        """Apply the "ge" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorWrappedPhiTensorPointer._apply_op(self, other, "__ge__")

    def __le__(
        self,
        other: Union[TensorWrappedPhiTensorPointer, MPCTensor, int, float, np.ndarray],
    ) -> Union[TensorWrappedPhiTensorPointer, MPCTensor]:
        """Apply the "le" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorWrappedPhiTensorPointer._apply_op(self, other, "__le__")

    def __eq__(  # type: ignore
        self,
        other: Union[TensorWrappedPhiTensorPointer, MPCTensor, int, float, np.ndarray],
    ) -> Union[TensorWrappedPhiTensorPointer, MPCTensor]:
        """Apply the "eq" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorWrappedPhiTensorPointer._apply_op(self, other, "__eq__")

    def __ne__(  # type: ignore
        self,
        other: Union[TensorWrappedPhiTensorPointer, MPCTensor, int, float, np.ndarray],
    ) -> Union[TensorWrappedPhiTensorPointer, MPCTensor]:
        """Apply the "ne" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorWrappedPhiTensorPointer._apply_op(self, other, "__ne__")

    def concatenate(
        self,
        other: TensorWrappedPhiTensorPointer,
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ) -> MPCTensor:
        """Apply the "add" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.


        Returns:
            Union[TensorWrappedPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        if not isinstance(other, TensorWrappedPhiTensorPointer):
            raise ValueError(
                f"Concatenate works only for TensorWrappedPhiTensorPointer got type: {type(other)}"
            )

        if self.client != other.client:

            parties = [self.client, other.client]

            self_mpc = MPCTensor(secret=self, shape=self.public_shape, parties=parties)
            other_mpc = MPCTensor(
                secret=other, shape=other.public_shape, parties=parties
            )

            return self_mpc.concatenate(other_mpc, *args, **kwargs)

        else:
            raise ValueError(
                "Concatenate method currently works only between two different clients."
            )

    def __truediv__(
        self,
        other: Union[TensorWrappedPhiTensorPointer, MPCTensor, int, float, np.ndarray],
    ) -> Union[TensorWrappedPhiTensorPointer, MPCTensor]:
        """Apply the "truediv" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return TensorWrappedPhiTensorPointer._apply_op(self, other, "__truediv__")

    def sum(
        self,
        *args: Tuple[Any, ...],
        **kwargs: Any,
    ) -> Union[
        TensorWrappedPhiTensorPointer, MPCTensor, TensorWrappedGammaTensorPointer
    ]:
        """Apply the "truediv" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        return self.gamma.sum(*args, **kwargs)

    def __getitem__(
        self, key: Union[int, bool, slice]
    ) -> TensorWrappedPhiTensorPointer:
        """Apply the slice  operation on "self"
        Args:
            y (Union[int,bool,slice]) : second operand.

        Returns:
            Union[TensorWrappedPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        attr_path_and_name = "syft.core.tensor.tensor.Tensor.__getitem__"
        result: TensorWrappedPhiTensorPointer
        min_vals = self.min_vals.__getitem__(key)
        max_vals = self.max_vals.__getitem__(key)

        result = TensorWrappedPhiTensorPointer(
            data_subjects=self.data_subjects,
            min_vals=min_vals,
            max_vals=max_vals,
            client=self.client,
        )

        # QUESTION can the id_at_location be None?
        result_id_at_location = getattr(result, "id_at_location", None)

        if result_id_at_location is not None:
            # first downcast anything primitive which is not already PyPrimitive
            (
                downcast_args,
                downcast_kwargs,
            ) = lib.python.util.downcast_args_and_kwargs(args=[key], kwargs={})

            # then we convert anything which isnt a pointer into a pointer
            pointer_args, pointer_kwargs = pointerize_args_and_kwargs(
                args=downcast_args,
                kwargs=downcast_kwargs,
                client=self.client,
                gc_enabled=False,
            )

            cmd = RunClassMethodAction(
                path=attr_path_and_name,
                _self=self,
                args=pointer_args,
                kwargs=pointer_kwargs,
                id_at_location=result_id_at_location,
                address=self.client.address,
            )
            self.client.send_immediate_msg_without_reply(msg=cmd)

        inherit_tags(
            attr_path_and_name=attr_path_and_name,
            result=result,
            self_obj=self,
            args=[key],
            kwargs={},
        )
        dummy_res = np.empty(self.public_shape).__getitem__(key)
        result.public_shape = dummy_res.shape
        result.public_dtype = self.public_dtype

        return result

    def ones_like(
        self,
        *args: Tuple[Any, ...],
        **kwargs: Any,
    ) -> TensorWrappedPhiTensorPointer:
        """Apply the "truediv" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        attr_path_and_name = "syft.core.tensor.tensor.Tensor.ones_like"
        result: TensorWrappedPhiTensorPointer
        min_vals = self.min_vals.ones_like(*args, **kwargs)
        max_vals = self.max_vals.ones_like(*args, **kwargs)

        result = TensorWrappedPhiTensorPointer(
            data_subjects=self.data_subjects,
            min_vals=min_vals,
            max_vals=max_vals,
            client=self.client,
        )

        # QUESTION can the id_at_location be None?
        result_id_at_location = getattr(result, "id_at_location", None)

        if result_id_at_location is not None:
            # first downcast anything primitive which is not already PyPrimitive
            (
                downcast_args,
                downcast_kwargs,
            ) = lib.python.util.downcast_args_and_kwargs(args=args, kwargs=kwargs)

            # then we convert anything which isnt a pointer into a pointer
            pointer_args, pointer_kwargs = pointerize_args_and_kwargs(
                args=downcast_args,
                kwargs=downcast_kwargs,
                client=self.client,
                gc_enabled=False,
            )

            cmd = RunClassMethodAction(
                path=attr_path_and_name,
                _self=self,
                args=pointer_args,
                kwargs=pointer_kwargs,
                id_at_location=result_id_at_location,
                address=self.client.address,
            )
            self.client.send_immediate_msg_without_reply(msg=cmd)

        inherit_tags(
            attr_path_and_name=attr_path_and_name,
            result=result,
            self_obj=self,
            args=[],
            kwargs={},
        )
        dummy_res = np.ones_like(np.empty(self.public_shape), *args, **kwargs)
        result.public_shape = dummy_res.shape
        result.public_dtype = self.public_dtype

        return result

    def exp(
        self,
    ) -> Union[TensorWrappedPhiTensorPointer, MPCTensor]:
        """Apply the "truediv" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        attr_path_and_name = "syft.core.tensor.tensor.Tensor.exp"

        # TODO: should modify to log reduction.
        def exp_reduction(val: np.ndarray) -> np.ndarray:
            pos_index = val >= 0
            neg_index = val < 0
            exp = np.exp((pos_index * val * -1) + (neg_index * val))
            pos_values = (pos_index) * exp
            neg_values = (neg_index) * exp * -1
            return pos_values + neg_values

        min_vals = self.min_vals.copy()
        min_vals.data = np.array(exp_reduction(min_vals.data))
        max_vals = self.max_vals.copy()
        max_vals.data = np.array(exp_reduction(max_vals.data))

        result = TensorWrappedPhiTensorPointer(
            data_subjects=self.data_subjects,
            min_vals=min_vals,
            max_vals=max_vals,
            client=self.client,
        )

        # QUESTION can the id_at_location be None?
        result_id_at_location = getattr(result, "id_at_location", None)

        if result_id_at_location is not None:
            # first downcast anything primitive which is not already PyPrimitive
            (
                downcast_args,
                downcast_kwargs,
            ) = lib.python.util.downcast_args_and_kwargs(args=[], kwargs={})

            # then we convert anything which isnt a pointer into a pointer
            pointer_args, pointer_kwargs = pointerize_args_and_kwargs(
                args=downcast_args,
                kwargs=downcast_kwargs,
                client=self.client,
                gc_enabled=False,
            )

            cmd = RunClassMethodAction(
                path=attr_path_and_name,
                _self=self,
                args=pointer_args,
                kwargs=pointer_kwargs,
                id_at_location=result_id_at_location,
                address=self.client.address,
            )
            self.client.send_immediate_msg_without_reply(msg=cmd)

        inherit_tags(
            attr_path_and_name=attr_path_and_name,
            result=result,
            self_obj=self,
            args=[],
            kwargs={},
        )

        result.public_shape = self.public_shape
        result.public_dtype = self.public_dtype

        return result

    def reciprocal(
        self,
    ) -> Union[TensorWrappedPhiTensorPointer, MPCTensor]:
        """Apply the "reciprocal" operation between "self" and "other"

        Args:
            y (Union[TensorWrappedPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        attr_path_and_name = "syft.core.tensor.tensor.Tensor.reciprocal"

        min_vals = self.min_vals.copy()
        min_vals.data = np.array(1 / min_vals.data)
        max_vals = self.max_vals.copy()
        max_vals.data = np.array(1 / max_vals.data)

        result = TensorWrappedPhiTensorPointer(
            data_subjects=self.data_subjects,
            min_vals=min_vals,
            max_vals=max_vals,
            client=self.client,
        )

        # QUESTION can the id_at_location be None?
        result_id_at_location = getattr(result, "id_at_location", None)

        if result_id_at_location is not None:
            # first downcast anything primitive which is not already PyPrimitive
            (
                downcast_args,
                downcast_kwargs,
            ) = lib.python.util.downcast_args_and_kwargs(args=[], kwargs={})

            # then we convert anything which isnt a pointer into a pointer
            pointer_args, pointer_kwargs = pointerize_args_and_kwargs(
                args=downcast_args,
                kwargs=downcast_kwargs,
                client=self.client,
                gc_enabled=False,
            )

            cmd = RunClassMethodAction(
                path=attr_path_and_name,
                _self=self,
                args=pointer_args,
                kwargs=pointer_kwargs,
                id_at_location=result_id_at_location,
                address=self.client.address,
            )
            self.client.send_immediate_msg_without_reply(msg=cmd)

        inherit_tags(
            attr_path_and_name=attr_path_and_name,
            result=result,
            self_obj=self,
            args=[],
            kwargs={},
        )

        result.public_shape = self.public_shape
        result.public_dtype = self.public_dtype

        return result

    def softmax(
        self,
    ) -> Union[TensorWrappedPhiTensorPointer, MPCTensor]:
        """Apply the "softmax" operation on "self"

        Args:
            y (Union[TensorWrappedPhiTensorPointer,MPCTensor,int,float,np.ndarray]) : second operand.

        Returns:
            Union[TensorWrappedPhiTensorPointer,MPCTensor] : Result of the operation.
        """
        attr_path_and_name = "syft.core.tensor.tensor.Tensor.softmax"

        # TODO: should modify to log reduction.
        def softmax(val: np.ndarray) -> np.ndarray:
            logits = val - val.max()
            numerator = np.exp(logits)
            inv = 1 / numerator.sum()
            return numerator * inv

        min_vals = self.min_vals.copy()
        min_vals.data = np.array(softmax(min_vals.data))
        max_vals = self.max_vals.copy()
        max_vals.data = np.array(softmax(max_vals.data))

        result = TensorWrappedPhiTensorPointer(
            data_subjects=self.data_subjects,
            min_vals=min_vals,
            max_vals=max_vals,
            client=self.client,
        )

        # QUESTION can the id_at_location be None?
        result_id_at_location = getattr(result, "id_at_location", None)

        if result_id_at_location is not None:
            # first downcast anything primitive which is not already PyPrimitive
            (
                downcast_args,
                downcast_kwargs,
            ) = lib.python.util.downcast_args_and_kwargs(args=[], kwargs={})

            # then we convert anything which isnt a pointer into a pointer
            pointer_args, pointer_kwargs = pointerize_args_and_kwargs(
                args=downcast_args,
                kwargs=downcast_kwargs,
                client=self.client,
                gc_enabled=False,
            )

            cmd = RunClassMethodAction(
                path=attr_path_and_name,
                _self=self,
                args=pointer_args,
                kwargs=pointer_kwargs,
                id_at_location=result_id_at_location,
                address=self.client.address,
            )
            self.client.send_immediate_msg_without_reply(msg=cmd)

        inherit_tags(
            attr_path_and_name=attr_path_and_name,
            result=result,
            self_obj=self,
            args=[],
            kwargs={},
        )

        result.public_shape = self.public_shape
        result.public_dtype = self.public_dtype

        return result

    @property
    def T(self) -> TensorWrappedPhiTensorPointer:
        # We always maintain a Tensor hierarchy Tensor ---> PT--> Actual Data
        attr_path_and_name = "syft.core.tensor.tensor.Tensor.T"

        result = TensorWrappedPhiTensorPointer(
            data_subjects=self.data_subjects,
            min_vals=self.min_vals.transpose(),
            max_vals=self.max_vals.transpose(),
            client=self.client,
        )

        # QUESTION can the id_at_location be None?
        result_id_at_location = getattr(result, "id_at_location", None)

        if result_id_at_location is not None:
            # first downcast anything primitive which is not already PyPrimitive
            (
                downcast_args,
                downcast_kwargs,
            ) = lib.python.util.downcast_args_and_kwargs(args=[], kwargs={})

            # then we convert anything which isnt a pointer into a pointer
            pointer_args, pointer_kwargs = pointerize_args_and_kwargs(
                args=downcast_args,
                kwargs=downcast_kwargs,
                client=self.client,
                gc_enabled=False,
            )

            cmd = GetOrSetPropertyAction(
                path=attr_path_and_name,
                id_at_location=result_id_at_location,
                address=self.client.address,
                _self=self,
                args=pointer_args,
                kwargs=pointer_kwargs,
                action=PropertyActions.GET,
                map_to_dyn=False,
            )
            self.client.send_immediate_msg_without_reply(msg=cmd)

        inherit_tags(
            attr_path_and_name=attr_path_and_name,
            result=result,
            self_obj=self,
            args=[],
            kwargs={},
        )

        result_public_shape = np.empty(self.public_shape).T.shape

        result.public_shape = result_public_shape
        result.public_dtype = self.public_dtype

        return result

    def one_hot(self: TensorWrappedPhiTensorPointer) -> np.array:
        tensor_size = np.empty(self.public_shape).size
        one_hot_Y = np.zeros((tensor_size, self.max_vals.data[0] + 1))
        one_hot_Y = one_hot_Y.T

        attr_path_and_name = "syft.core.tensor.tensor.Tensor.one_hot"

        result = TensorWrappedPhiTensorPointer(
            data_subjects=self.data_subjects,
            min_vals=self.min_vals,
            max_vals=self.max_vals,
            client=self.client,
        )

        # QUESTION can the id_at_location be None?
        result_id_at_location = getattr(result, "id_at_location", None)

        if result_id_at_location is not None:
            # first downcast anything primitive which is not already PyPrimitive
            (
                downcast_args,
                downcast_kwargs,
            ) = lib.python.util.downcast_args_and_kwargs(args=[], kwargs={})

            # then we convert anything which isnt a pointer into a pointer
            pointer_args, pointer_kwargs = pointerize_args_and_kwargs(
                args=downcast_args,
                kwargs=downcast_kwargs,
                client=self.client,
                gc_enabled=False,
            )

            cmd = RunClassMethodAction(
                path=attr_path_and_name,
                _self=self,
                args=pointer_args,
                kwargs=pointer_kwargs,
                id_at_location=result_id_at_location,
                address=self.client.address,
            )
            self.client.send_immediate_msg_without_reply(msg=cmd)

        inherit_tags(
            attr_path_and_name=attr_path_and_name,
            result=result,
            self_obj=self,
            args=[],
            kwargs={},
        )

        result.public_shape = one_hot_Y.shape
        result.public_dtype = self.public_dtype

        return result

    def to_local_object_without_private_data_child(self) -> PhiTensor:
        """Convert this pointer into a partial version of the PhiTensor but without
        any of the private data therein."""
        # relative
        from ..tensor import Tensor

        public_shape = getattr(self, "public_shape", None)
        public_dtype = getattr(self, "public_dtype", None)
        return Tensor(
            child=PhiTensor(
                child=FixedPrecisionTensor(value=np.empty(self.data_subjects.shape)),
                data_subjects=self.data_subjects,
                min_vals=self.min_vals,  # type: ignore
                max_vals=self.max_vals,  # type: ignore
            ),
            public_shape=public_shape,
            public_dtype=public_dtype,
        )


@implements(TensorWrappedPhiTensorPointer, np.ones_like)
def ones_like(
    tensor: TensorWrappedPhiTensorPointer,
    *args: Tuple[Any, ...],
    **kwargs: Dict[Any, Any],
) -> TensorWrappedPhiTensorPointer:
    return tensor.ones_like(*args, **kwargs)


@serializable(capnp_bytes=True)
class PhiTensor(PassthroughTensor, ADPTensor):
    PointerClassOverride = TensorWrappedPhiTensorPointer
    # __attr_allowlist__ = ["child", "min_vals", "max_vals", "data_subjects"]
    __slots__ = (
        "child",
        "min_vals",
        "max_vals",
        "data_subjects",
    )

    def __init__(
        self,
        child: Union[Sequence, NDArray],
        data_subjects: Union[DataSubjectArray, NDArray],
        min_vals: Union[np.ndarray, lazyrepeatarray],
        max_vals: Union[np.ndarray, lazyrepeatarray],
    ) -> None:
        # self.data_subjects: Union[DataSubjectList, np.ndarray]
        # child = the actual private data
        super().__init__(child)

        # lazyrepeatarray matching the shape of child
        if not isinstance(min_vals, lazyrepeatarray):
            min_vals = lazyrepeatarray(data=min_vals, shape=child.shape)  # type: ignore
        if not isinstance(max_vals, lazyrepeatarray):
            max_vals = lazyrepeatarray(data=max_vals, shape=child.shape)  # type: ignore
        self.min_vals = min_vals
        self.max_vals = max_vals

        numpy_data_subjects: np.ndarray = DataSubjectArray.from_objs(data_subjects)
        self.data_subjects = numpy_data_subjects
        if numpy_data_subjects.shape != self.shape:
            raise ValueError(
                f"DataSubjects shape: {numpy_data_subjects.shape} should match data shape: {self.shape}"
            )

    @property
    def proxy_public_kwargs(self) -> Dict[str, Any]:
        return {
            "min_vals": self.min_vals,
            "max_vals": self.max_vals,
            "data_subjects": self.data_subjects,
        }

    # def init_pointer(
    #     self,
    #     client: Any,
    #     id_at_location: Optional[UID] = None,
    #     object_type: str = "",
    #     tags: Optional[List[str]] = None,
    #     description: str = "",
    # ) -> TensorWrappedPhiTensorPointer:
    #     return TensorWrappedPhiTensorPointer(
    #         # Arguments specifically for SEPhiTensor
    #         data_subjects=self.data_subjects,
    #         min_vals=self.min_vals,
    #         max_vals=self.max_vals,
    #         # Arguments required for a Pointer to work
    #         client=client,
    #         id_at_location=id_at_location,
    #         object_type=object_type,
    #         tags=tags,
    #         description=description,
    #     )

    @property
    def gamma(self) -> GammaTensor:
        """Property to cast this tensor into a GammaTensor"""
        return self.create_gamma()

    def copy(self, order: Optional[str] = "K") -> PhiTensor:
        """Return copy of the given object"""

        return PhiTensor(
            child=self.child.copy(order=order),
            min_vals=self.min_vals.copy(order=order),
            max_vals=self.max_vals.copy(order=order),
            data_subjects=self.data_subjects.copy(),
        )

    def all(self) -> bool:
        return self.child.all()

    def any(self) -> bool:
        return self.child.any()

    def copy_with(self, child: np.ndarray) -> PhiTensor:
        new_tensor = self.copy()
        new_tensor.child = child
        return new_tensor

    def __getitem__(self, item: Union[str, int, slice, PassthroughTensor]) -> PhiTensor:
        if isinstance(item, PassthroughTensor):
            data = self.child[item.child]
            return PhiTensor(
                child=data,
                min_vals=lazyrepeatarray(data=data, shape=data.shape),
                max_vals=lazyrepeatarray(data=data, shape=data.shape),
                data_subjects=self.data_subjects[item.child],
            )
        else:
            data = self.child[item]
            return PhiTensor(
                child=data,
                min_vals=lazyrepeatarray(data=data, shape=data.shape),
                max_vals=lazyrepeatarray(data=data, shape=data.shape),
                data_subjects=self.data_subjects[item],
            )

    def zeros_like(
        self,
        *args: Tuple[Any, ...],
        **kwargs: Any,
    ) -> Union[PhiTensor, GammaTensor]:
        # TODO: Add support for axes arguments later
        min_vals = self.min_vals.zeros_like(*args, **kwargs)
        max_vals = self.max_vals.zeros_like(*args, **kwargs)

        child = (
            np.zeros_like(self.child, *args, **kwargs)
            if isinstance(self.child, np.ndarray)
            else self.child.ones_like(*args, **kwargs)
        )

        return PhiTensor(
            child=child,
            min_vals=min_vals,
            max_vals=max_vals,
            data_subjects=self.data_subjects,
        )

    def __setitem__(
        self,
        key: Union[int, slice, NDArray],
        value: Union[PhiTensor, GammaTensor, np.ndarray],
    ) -> Union[PhiTensor, GammaTensor]:
        if isinstance(value, PhiTensor):
            self.child[key] = value.child
            minv = value.child.min()
            maxv = value.child.max()

            if minv < self.min_vals.data.min():
                self.min_vals.data = minv

            if maxv > self.max_vals.data.max():
                self.max_vals.data = maxv

            gamma_output = self.gamma
            gamma_output[key] = value.gamma
            # print("It's on the right track")
            return gamma_output

        elif isinstance(value, GammaTensor):
            gamma = self.gamma
            gamma[key] = value
            return gamma
        elif isinstance(value, np.ndarray):
            self.child[key] = value
            minv = value.min()
            maxv = value.max()

            if minv < self.min_vals.data.min():
                self.min_vals.data = minv

            if maxv > self.max_vals.data.max():
                self.max_vals.data = maxv

            return PhiTensor(
                child=self.child,
                data_subjects=self.data_subjects,
                min_vals=self.min_vals,
                max_vals=self.max_vals,
            )
        else:
            raise NotImplementedError

    def abs(self) -> PhiTensor:
        data = self.child
        output = np.abs(data)

        return PhiTensor(
            child=output,
            data_subjects=self.data_subjects,
            min_vals=np.abs(self.min_vals.data),
            max_vals=np.abs(self.min_vals.data),
        )

    def reshape(self, *shape: Tuple[int, ...]) -> PhiTensor:

        data = self.child
        output_data = np.reshape(data, *shape)

        return PhiTensor(
            child=output_data,
            data_subjects=np.reshape(self.data_subjects, *shape),
            min_vals=output_data.min(),
            max_vals=output_data.max(),
        )

    def pad(self, width: int, padding_mode: str = "reflect") -> PhiTensor:
        data = self.child

        if padding_mode == "reflect":
            pad_left = pad_right = pad_top = pad_bottom = width
            # RGB image
            if len(data.shape) == 3:
                output_data = np.pad(
                    data,
                    ((pad_top, pad_bottom), (pad_left, pad_right), (0, 0)),
                    padding_mode,
                )
                output_data_subjects = np.pad(
                    self.data_subjects,
                    ((pad_top, pad_bottom), (pad_left, pad_right), (0, 0)),
                    padding_mode,
                )
            # Grayscale image
            elif len(data.shape) == 2:
                output_data = np.pad(
                    data, ((pad_top, pad_bottom), (pad_left, pad_right)), padding_mode
                )
                output_data_subjects = np.pad(
                    self.data_subjects,
                    ((pad_top, pad_bottom), (pad_left, pad_right)),
                    padding_mode,
                )
            else:
                output_data = np.pad(data, width, padding_mode)
                output_data_subjects = np.pad(self.data_subjects, width, padding_mode)
        else:
            raise NotImplementedError

        output_min_val, output_max_val = output_data.min(), output_data.max()
        return PhiTensor(
            child=output_data,
            data_subjects=output_data_subjects,
            min_vals=output_min_val,
            max_vals=output_max_val,
        )

    def ravel(self) -> PhiTensor:
        data = self.child
        output_data = data.ravel()

        output_data_subjects = self.data_subjects.ravel()

        min_vals = lazyrepeatarray(data=self.min_vals.data, shape=output_data.shape)
        max_vals = lazyrepeatarray(data=self.max_vals.data, shape=output_data.shape)

        return PhiTensor(
            child=output_data,
            data_subjects=output_data_subjects,
            min_vals=min_vals,
            max_vals=max_vals,
        )

    def random_horizontal_flip(self, p: float = 0.5) -> PhiTensor:
        """Could make more efficient by not encoding/decoding FPT"""
        if np.random.random() <= p:
            return PhiTensor(
                child=np.fliplr(self.child),
                data_subjects=self.data_subjects,
                min_vals=self.min_vals.horizontal_flip(),
                max_vals=self.max_vals.horizontal_flip(),
            )
        else:
            return self

    def random_vertical_flip(self, p: float = 0.5) -> PhiTensor:
        """Could make more efficient by not encoding/decoding FPT"""
        if np.random.random() <= p:
            return PhiTensor(
                child=np.flipud(self.child),
                data_subjects=self.data_subjects,
                min_vals=self.min_vals.vertical_flip(),
                max_vals=self.max_vals.vertical_flip(),
            )
        else:
            return self

    def random_rotation(self, degrees: Union[int, Tuple]) -> PhiTensor:
        if isinstance(degrees, int):
            angle = np.random.randint(low=-degrees, high=degrees)
        elif isinstance(degrees, tuple):
            angle = np.random.randint(low=degrees[0], high=degrees[1])

        rotated_data_value = rotate(self.child, angle)

        return PhiTensor(
            child=rotated_data_value,
            data_subjects=self.data_subjects,
            min_vals=rotated_data_value.min(),
            max_vals=rotated_data_value.max(),
        )

    def max(self, axis: Optional[Union[int, Tuple[int, ...]]] = None) -> PhiTensor:
        indices = self.child.argmax(axis)
        result = self.child.max(axis)
        return PhiTensor(
            child=result,
            data_subjects=self.data_subjects[indices],
            min_vals=lazyrepeatarray(data=result.min(), shape=result.shape),
            max_vals=lazyrepeatarray(data=result.max(), shape=result.shape),
        )

    def _argmax(self, axis: Optional[int]) -> PhiTensor:
        return self.child.argmax(axis)

    def unravel_argmax(
        self, axis: Optional[int] = None
    ) -> Tuple[np.ndarray]:  # possible privacy violation?
        arg_result = self._argmax(axis=axis)
        shape = self.shape
        return np.unravel_index(arg_result, shape)

    def mean(
        self,
        axis: Optional[Union[int, Tuple[int, ...]]] = None,
        **kwargs: Any,
    ) -> PhiTensor:
        result = self.child.mean(axis, **kwargs)

        return PhiTensor(
            child=result,
            data_subjects=self.data_subjects.mean(axis, **kwargs),
            min_vals=lazyrepeatarray(data=self.min_vals.data, shape=result.shape),
            max_vals=lazyrepeatarray(data=self.max_vals.data, shape=result.shape),
        )

    def std(
        self,
        axis: Optional[Union[int, Tuple[int, ...]]] = None,
        **kwargs: Any,
    ) -> PhiTensor:
        result = self.child.std(axis, **kwargs)
        return PhiTensor(
            child=result,
            data_subjects=self.data_subjects.std(axis, **kwargs),
            min_vals=lazyrepeatarray(data=0, shape=result.shape),
            max_vals=lazyrepeatarray(
                data=0.25
                * (self.max_vals.data - self.min_vals.data)
                ** 2,  # rough approximation, could be off
                shape=result.shape,
            ),
        )

    def sqrt(self) -> PhiTensor:
        result = np.sqrt(self.child)
        return PhiTensor(
            child=result,
            data_subjects=np.sqrt(self.data_subjects),
            min_vals=lazyrepeatarray(
                data=np.sqrt(self.min_vals.data), shape=result.shape
            ),
            max_vals=lazyrepeatarray(
                data=np.sqrt(self.max_vals.data), shape=result.shape
            ),
        )

    def normalize(
        self, mean: Union[float, Sequence[float]], std: Union[float, Sequence[float]]
    ) -> PhiTensor:
        # TODO: Double check if normalization bounds are correct; they might be data dependent
        if isinstance(mean, float) and isinstance(std, float):
            return PhiTensor(
                child=(self.child - mean) / std,
                data_subjects=self.data_subjects,
                min_vals=(self.min_vals - mean) * (1 / std),
                max_vals=(self.max_vals - mean) * (1 / std),
            )
        else:
            # This is easily doable in the future
            raise NotImplementedError

    def create_gamma(self) -> GammaTensor:
        """Return a new Gamma tensor based on this phi tensor"""
        gamma_tensor = GammaTensor(
            child=self.child,
            data_subjects=self.data_subjects,
            min_vals=self.min_vals,
            max_vals=self.max_vals,
        )

        return gamma_tensor

    def view(self, *args: List[Any]) -> PhiTensor:
        # TODO: Figure out how to fix lazyrepeatarray reshape

        data = self.child.reshape(*args)
        return PhiTensor(
            child=data,
            data_subjects=self.data_subjects,
            min_vals=lazyrepeatarray(data=self.min_vals.data.min(), shape=data.shape),
            max_vals=lazyrepeatarray(data=self.max_vals.data.max(), shape=data.shape),
        )

    def publish(
        self,
        get_budget_for_user: Callable,
        deduct_epsilon_for_user: Callable,
        ledger: DataSubjectLedger,
        sigma: float,
    ) -> AcceptableSimpleType:
        print("PUBLISHING TO GAMMA:")
        print(self.child)

        gamma = self.gamma
        # gamma.func = lambda x: x
        gamma.state[gamma.id] = gamma

        res = gamma.publish(
            get_budget_for_user=get_budget_for_user,
            deduct_epsilon_for_user=deduct_epsilon_for_user,
            ledger=ledger,
            sigma=sigma,
        )

        print("Final Values", res)

        return res

    @property
    def value(self) -> np.ndarray:
        return self.child

    def astype(self, np_type: np.dtype) -> PhiTensor:
        return self.__class__(
            child=self.child.astype(np_type),
            data_subjects=self.data_subjects,
            min_vals=self.min_vals.astype(np_type),
            max_vals=self.max_vals.astype(np_type),
            # scalar_manager=self.scalar_manager,
        )

    @property
    def shape(self) -> Tuple[Any, ...]:
        if self.child is None:
            return ()
        else:
            return self.child.shape

    def __repr__(self) -> str:
        """Pretty print some information, optimized for Jupyter notebook viewing."""
        return (
            f"{self.__class__.__name__}(child={self.child}, "
            + f"min_vals={self.min_vals}, max_vals={self.max_vals})"
        )

    def __eq__(self, other: Any) -> Union[PhiTensor, GammaTensor]:  # type: ignore
        # TODO: what about data_subjects and min / max values?
        if is_acceptable_simple_type(other) or len(self.child) == len(other.child):
            gamma_output = False
            if is_acceptable_simple_type(other):
                result = self.child == other
            else:
                # check data_subjects match, if they dont gamma_output = True
                result = self.child == other.child
                if isinstance(result, GammaTensor):  # TODO: Check this
                    gamma_output = True
            if not gamma_output:
                # min_vals=self.min_vals * 0.0,
                # max_vals=self.max_vals * 0.0 + 1.0,
                return self.copy_with(child=result)
            else:
                return self.copy_with(child=result).gamma
        else:
            raise Exception(
                "Tensor dims do not match for __eq__: "
                + f"{len(self.child)} != {len(other.child)}"
            )

    def __add__(self, other: SupportedChainType) -> Union[PhiTensor, GammaTensor]:

        # if the tensor being added is also private
        if isinstance(other, PhiTensor):
            return self.gamma + other.gamma
            # if self.data_subjects != other.data_subjects:
            #     return self.gamma + other.gamma

            # return PhiTensor(
            #     child=self.child + other.child,
            #     min_vals=self.min_vals + other.min_vals,
            #     max_vals=self.max_vals + other.max_vals,
            #     data_subjects=self.data_subjects,
            #     # scalar_manager=self.scalar_manager,
            # )

        # if the tensor being added is a public tensor / int / float / etc.
        elif is_acceptable_simple_type(other):

            return PhiTensor(
                child=self.child + other,
                min_vals=self.min_vals + other,
                max_vals=self.max_vals + other,
                data_subjects=self.data_subjects,
            )

        elif isinstance(other, GammaTensor):
            return self.gamma + other
        else:
            print("Type is unsupported:" + str(type(other)))
            raise NotImplementedError

    def __sub__(self, other: SupportedChainType) -> Union[PhiTensor, GammaTensor]:

        if isinstance(other, PhiTensor):
            return self.gamma - other.gamma
            # diff_data_subjects = (
            #     self.data_subjects.one_hot_lookup != other.data_subjects.one_hot_lookup
            # )
            # diff_data_subjects = (
            #     diff_data_subjects
            #     if isinstance(diff_data_subjects, bool)
            #     else diff_data_subjects.any()
            # )
            # if diff_data_subjects:
            #     return self.gamma - other.gamma
            #     # raise NotImplementedError

            # data = self.child - other.child
            # min_min = self.min_vals.data - other.min_vals.data
            # min_max = self.min_vals.data - other.max_vals.data
            # max_min = self.max_vals.data - other.min_vals.data
            # max_max = self.max_vals.data - other.max_vals.data
            # _min_vals = np.minimum.reduce([min_min, min_max, max_min, max_max])
            # _max_vals = np.maximum.reduce([min_min, min_max, max_min, max_max])
            # min_vals = self.min_vals.copy()
            # min_vals.data = _min_vals
            # max_vals = self.max_vals.copy()
            # max_vals.data = _max_vals

            # data_subjects = self.data_subjects

        elif is_acceptable_simple_type(other):
            if isinstance(other, np.ndarray):
                if not is_broadcastable(other.shape, self.child.shape):  # type: ignore
                    raise Exception(
                        f"Shapes do not match for subtraction: {self.child.shape} and {other.shape}"
                    )
            data = self.child - other
            min_vals = self.min_vals - other
            max_vals = self.max_vals - other
            data_subjects = self.data_subjects
        elif isinstance(other, GammaTensor):
            return self.gamma - other
        else:
            print("Type is unsupported:" + str(type(other)))
            raise NotImplementedError
        return PhiTensor(
            child=data,
            data_subjects=data_subjects,
            min_vals=min_vals,
            max_vals=max_vals,
        )

    def __mul__(self, other: SupportedChainType) -> Union[PhiTensor, GammaTensor]:

        if isinstance(other, PhiTensor):
            return self.gamma + other.gamma
            # if self.data_subjects != other.data_subjects:
            #     return self.gamma * other.gamma

            # data = self.child * other.child

            # min_min = self.min_vals.data * other.min_vals.data
            # min_max = self.min_vals.data * other.max_vals.data
            # max_min = self.max_vals.data * other.min_vals.data
            # max_max = self.max_vals.data * other.max_vals.data

            # _min_vals = np.min([min_min, min_max, max_min, max_max], axis=0)  # type: ignore
            # _max_vals = np.max([min_min, min_max, max_min, max_max], axis=0)  # type: ignore
            # min_vals = self.min_vals.copy()
            # min_vals.data = _min_vals
            # max_vals = self.max_vals.copy()
            # max_vals.data = _max_vals

            # data_subjects = self.data_subjects

            # return PhiTensor(
            #     child=data,
            #     data_subjects=data_subjects,
            #     min_vals=min_vals,
            #     max_vals=max_vals,
            # )
        elif is_acceptable_simple_type(other):

            data = self.child * other

            min_min = self.min_vals.data * other
            min_max = self.min_vals.data * other
            max_min = self.max_vals.data * other
            max_max = self.max_vals.data * other

            _min_vals = np.min([min_min, min_max, max_min, max_max], axis=0)  # type: ignore
            _max_vals = np.max([min_min, min_max, max_min, max_max], axis=0)  # type: ignore
            min_vals = self.min_vals.copy()
            min_vals.data = _min_vals
            max_vals = self.max_vals.copy()
            max_vals.data = _max_vals

            data_subjects = self.data_subjects

            return PhiTensor(
                child=data,
                data_subjects=data_subjects,
                min_vals=min_vals,
                max_vals=max_vals,
            )
        elif isinstance(other, GammaTensor):
            return self.gamma * other
        else:
            print("Type is unsupported:" + str(type(other)))
            raise NotImplementedError

    def __rtruediv__(self, other: SupportedChainType) -> Union[PhiTensor, GammaTensor]:

        if is_acceptable_simple_type(other):
            return PhiTensor(
                child=(1 / self.child) * other,
                min_vals=(1 / self.min_vals) * other,
                max_vals=(1 / self.max_vals) * other,
                data_subjects=self.data_subjects,
                # scalar_manager=self.scalar_manager,
            )

        elif isinstance(other, GammaTensor):
            return (1 / self.gamma) * other
        else:
            print("Type is unsupported:" + str(type(other)))
            raise NotImplementedError

    def __matmul__(
        self, other: Union[np.ndarray, PhiTensor]
    ) -> Union[PhiTensor, GammaTensor]:
        if not isinstance(other, (np.ndarray, PhiTensor, GammaTensor)):
            raise Exception(
                f"Matrix multiplication not yet implemented for type {type(other)}"
            )
        else:
            # Modify before merge, to know is broadcast is actually necessary
            if False:  # and not is_broadcastable(self.shape, other.shape):
                raise Exception(
                    f"Shapes not broadcastable: {self.shape} and {other.shape}"
                )
            else:
                if isinstance(other, np.ndarray):
                    data = self.child.__matmul__(other)
                    min_vals = self.min_vals.__matmul__(other)
                    max_vals = self.max_vals.__matmul__(other)
                    output_ds = self.data_subjects @ other
                elif isinstance(other, PhiTensor):
                    return self.gamma @ other.gamma
                    # if self.data_subjects != other.data_subjects:
                    #     return self.gamma @ other.gamma
                    # else:
                    #     data = self.child.__matmul__(other.child)
                    #     min_vals = self.min_vals.__matmul__(other.min_vals)
                    #     max_vals = self.max_vals.__matmul__(other.max_vals)
                    #     output_ds = DataSubjectList(
                    #         one_hot_lookup=np.concatenate(
                    #             (
                    #                 self.data_subjects.one_hot_lookup,
                    #                 other.data_subjects.one_hot_lookup,
                    #             )
                    #         ),
                    #         data_subjects_indexed=np.concatenate(
                    #             (np.zeros_like(data), np.ones_like(data))
                    #         ),  # replace with (1, *data.shape) if inc shape
                    #     )

                elif isinstance(other, GammaTensor):
                    return self.gamma @ other
                else:
                    print("Type is unsupported:" + str(type(other)))
                    raise NotImplementedError

                return PhiTensor(
                    child=data,
                    max_vals=max_vals,
                    min_vals=min_vals,
                    data_subjects=output_ds,
                )

    def __rmatmul__(
        self, other: Union[np.ndarray, PhiTensor]
    ) -> Union[PhiTensor, GammaTensor]:
        if not isinstance(other, (np.ndarray, PhiTensor, GammaTensor)):
            raise Exception(
                f"Matrix multiplication not yet implemented for type {type(other)}"
            )
        else:
            # Modify before merge, to know is broadcast is actually necessary
            if False:  # and not is_broadcastable(self.shape, other.shape):
                raise Exception(
                    f"Shapes not broadcastable: {self.shape} and {other.shape}"
                )
            else:
                if isinstance(other, np.ndarray):
                    data = self.child.__rmatmul__(other)
                    min_vals = self.min_vals.__rmatmul__(other)
                    max_vals = self.max_vals.__rmatmul__(other)
                    output_ds = self.data_subjects.__rmatmul__(other)
                elif isinstance(other, PhiTensor):
                    return self.gamma.__rmatmul__(other.gamma)
                    # if self.data_subjects != other.data_subjects:
                    #     # return convert_to_gamma_tensor(self).__matmul__(convert_to_gamma_tensor(other))
                    #     raise NotImplementedError
                    # else:
                    #     data = self.child.__rmatmul__(other.child)
                    #     # _min_vals = np.array(
                    #     #     [self.min_vals.data.__matmul__(other.min_vals.data)]
                    #     # )
                    #     # _max_vals = np.array(
                    #     #     [self.max_vals.data.__matmul__(other.max_vals.data)]
                    #     # )
                    #     # min_vals = self.min_vals.copy()
                    #     # min_vals.data = _min_vals
                    #     # max_vals = self.max_vals.copy()
                    #     # max_vals.data = _max_vals
                    #     min_vals = self.min_vals.__rmatmul__(other.min_vals)
                    #     max_vals = self.max_vals.__rmatmul__(other.max_vals)

                else:
                    print("Type is unsupported:" + str(type(other)))
                    raise NotImplementedError

                return PhiTensor(
                    child=data,
                    max_vals=max_vals,
                    min_vals=min_vals,
                    data_subjects=output_ds,
                )

    def clip(self, a_min: float, a_max: float) -> PhiTensor:
        output_data = np.clip(self.child, a_min, a_max)

        min_v = np.clip(self.min_vals.data, a_min, a_max)
        max_v = np.clip(self.max_vals.data, a_min, a_max)

        min_vals = lazyrepeatarray(data=min_v, shape=output_data.shape)
        max_vals = lazyrepeatarray(data=max_v, shape=output_data.shape)

        return PhiTensor(
            child=output_data,
            data_subjects=self.data_subjects,
            min_vals=min_vals,
            max_vals=max_vals,
        )

    def transpose(self, *args: Any, **kwargs: Any) -> PhiTensor:
        """Transposes self.child, min_vals, and max_vals if these can be transposed, otherwise doesn't change them."""
        data: np.ndarray
        if (
            isinstance(self.child, int)
            or isinstance(self.child, float)
            or isinstance(self.child, bool)
        ):
            # For these data types, the transpose operation is meaningless, so don't change them.
            data = self.child  # type: ignore
            print(
                f"Warning: Tensor data was of type {type(data)}, transpose operation had no effect."
            )
        else:
            data = self.child.transpose(*args)

        # TODO: Should we give warnings for min_vals and max_vals being single floats/integers/booleans too?
        if (
            isinstance(self.min_vals, int)
            or isinstance(self.min_vals, float)
            or isinstance(self.min_vals, bool)
        ):
            # For these data types, the transpose operation is meaningless, so don't change them.
            min_vals = self.min_vals
            # print(f'Warning: Tensor data was of type {type(data)}, transpose operation had no effect.')
        else:
            min_vals = data.min()

        if (
            isinstance(self.max_vals, int)
            or isinstance(self.max_vals, float)
            or isinstance(self.max_vals, bool)
        ):
            # For these data types, the transpose operation is meaningless, so don't change them.
            max_vals = self.max_vals
            # print(f'Warning: Tensor data was of type {type(data)}, transpose operation had no effect.')
        else:
            max_vals = data.max()

        output_ds = self.data_subjects.transpose(*args)

        return PhiTensor(
            child=data,
            data_subjects=output_ds,
            min_vals=min_vals,
            max_vals=max_vals,
        )

    def concatenate(
        self,
        other: Union[np.ndarray, PhiTensor],
        *args: List[Any],
        **kwargs: Dict[str, Any],
    ) -> Union[PhiTensor, GammaTensor]:

        # if the tensor being added is also private
        if isinstance(other, PhiTensor):
            if self.data_subjects != other.data_subjects:
                return self.gamma + other.gamma

            return PhiTensor(
                child=self.child.concatenate(other.child, *args, **kwargs),
                min_vals=self.min_vals.concatenate(other.min_vals, *args, **kwargs),
                max_vals=self.max_vals.concatenate(other.max_vals, *args, **kwargs),
                data_subjects=self.data_subjects,
            )

        elif is_acceptable_simple_type(other):
            raise NotImplementedError
        else:
            print("Type is unsupported:" + str(type(other)))
            raise NotImplementedError

    def __lt__(self, other: SupportedChainType) -> Union[PhiTensor, GammaTensor]:

        # if the tensor being compared is also private
        if isinstance(other, PhiTensor):
            return self.gamma.__lt__(other.gamma)

            # if self.data_subjects != other.data_subjects:
            #     # return self.gamma < other.gamma
            #     raise NotImplementedError

            # if len(self.child) != len(other.child):
            #     raise Exception(
            #         f"Tensor dims do not match for __lt__: {len(self.child)} != {len(other.child)}"  # type: ignore
            #     )

            # data = (
            #     self.child < other.child
            # )  # the * 1 just makes sure it returns integers instead of True/False
            # min_vals = self.min_vals * 0
            # max_vals = (self.max_vals * 0) + 1
            # data_subjects = self.data_subjects

            # return PhiTensor(
            #     child=data,
            #     data_subjects=data_subjects,
            #     min_vals=min_vals,
            #     max_vals=max_vals,
            # )

        # if the tensor being compared is a public tensor / int / float / etc.
        elif is_acceptable_simple_type(other):

            data = self.child < other
            min_vals = self.min_vals * 0
            max_vals = (self.max_vals * 0) + 1
            data_subjects = self.data_subjects

            return PhiTensor(
                child=data,
                data_subjects=data_subjects,
                min_vals=min_vals,
                max_vals=max_vals,
            )

        else:
            return NotImplementedError  # type: ignore

    def __le__(self, other: SupportedChainType) -> Union[PhiTensor, GammaTensor]:

        # if the tensor being compared is also private
        if isinstance(other, PhiTensor):
            return self.gamma.__le__(other.gamma)

            # if self.data_subjects != other.data_subjects:
            #     # return self.gamma < other.gamma
            #     raise NotImplementedError

            # if len(self.child) != len(other.child):
            #     raise Exception(
            #         f"Tensor dims do not match for __le__: {len(self.child)} != {len(other.child)}"  # type: ignore
            #     )

            # data = (
            #     self.child <= other.child
            # )  # the * 1 just makes sure it returns integers instead of True/False
            # min_vals = self.min_vals * 0
            # max_vals = (self.max_vals * 0) + 1
            # data_subjects = self.data_subjects

            # return PhiTensor(
            #     child=data,
            #     data_subjects=data_subjects,
            #     min_vals=min_vals,
            #     max_vals=max_vals,
            # )

        # if the tensor being compared is a public tensor / int / float / etc.
        elif is_acceptable_simple_type(other):

            data = self.child <= other
            min_vals = self.min_vals * 0
            max_vals = (self.max_vals * 0) + 1
            data_subjects = self.data_subjects

            return PhiTensor(
                child=data,
                data_subjects=data_subjects,
                min_vals=min_vals,
                max_vals=max_vals,
            )

        else:
            return NotImplementedError  # type: ignore

    def __gt__(self, other: SupportedChainType) -> Union[PhiTensor, GammaTensor]:

        # if the tensor being compared is also private
        if isinstance(other, PhiTensor):
            return self.gamma.__gt__(other.gamma)

            # if self.data_subjects != other.data_subjects:
            #     # return self.gamma < other.gamma
            #     raise NotImplementedError

            # if len(self.child) != len(other.child):
            #     raise Exception(
            #         f"Tensor dims do not match for __gt__: {len(self.child)} != {len(other.child)}"  # type: ignore
            #     )

            # data = (
            #     self.child > other.child
            # )  # the * 1 just makes sure it returns integers instead of True/False
            # min_vals = self.min_vals * 0
            # max_vals = (self.max_vals * 0) + 1
            # data_subjects = self.data_subjects

            # return PhiTensor(
            #     child=data,
            #     data_subjects=data_subjects,
            #     min_vals=min_vals,
            #     max_vals=max_vals,
            # )

        # if the tensor being compared is a public tensor / int / float / etc.
        elif is_acceptable_simple_type(other):

            data = self.child > other
            min_vals = self.min_vals * 0
            max_vals = (self.max_vals * 0) + 1
            data_subjects = self.data_subjects

            return PhiTensor(
                child=data,
                data_subjects=data_subjects,
                min_vals=min_vals,
                max_vals=max_vals,
            )
        else:
            raise NotImplementedError  # type: ignore

    # Re enable after testing
    def dot(
        self, other: Union[PhiTensor, GammaTensor, np.ndarray]
    ) -> Union[PhiTensor, GammaTensor]:
        if isinstance(other, np.ndarray):
            return PhiTensor(
                child=np.dot(self.child, other),
                min_vals=np.dot(self.min_vals, other),
                max_vals=np.dot(self.max_vals, other),
                data_subjects=np.dot(self.data_subjects, other),
            )
        elif isinstance(other, PhiTensor):
            return self.gamma.dot(other.gamma)
            # if self.data_subjects.one_hot_lookup == other.data_subjects.one_hot_lookup:
            #     return PhiTensor(
            #         child=np.dot(self.child, other.child),
            #         min_vals=np.dot(self.min_vals, other.min_vals),
            #         max_vals=np.dot(self.max_vals, other.max_vals),
            #         data_subjects=self.data_subjects,
            #     )
            # else:
            #     return self.gamma.dot(other.gamma)
        elif isinstance(other, GammaTensor):
            return self.gamma.dot(other)
        else:
            raise NotImplementedError

    def sum(
        self,
        axis: Optional[Union[int, Tuple[int, ...]]] = None,
        **kwargs: Any,
    ) -> Union[PhiTensor, GammaTensor]:
        return self.gamma.sum(axis, **kwargs)
        # # TODO: Add support for axes arguments later
        # min_val = self.min_vals.sum(axis=axis)
        # max_val = self.max_vals.sum(axis=axis)
        # if len(self.data_subjects.one_hot_lookup) == 1:
        #     result = self.child.sum(axis=axis)
        #     return PhiTensor(
        #         child=result,
        #         min_vals=min_val,
        #         max_vals=max_val,
        #         data_subjects=self.data_subjects.sum(target_shape=result.shape),
        #     )
        # result = self.child.sum(axis=axis)
        # return GammaTensor(
        #     child=result,
        #     data_subjects=self.data_subjects.sum(target_shape=result.shape),
        #     min_vals=min_val,
        #     max_vals=max_val,
        # )

    def expand_dims(self, axis: int) -> PhiTensor:
        result = np.expand_dims(self.child, axis=axis)
        minv = self.min_vals.copy()
        minv.shape = result.shape
        maxv = self.max_vals.copy()
        maxv.shape = result.shape

        return PhiTensor(
            child=result,
            min_vals=minv,
            max_vals=maxv,
            data_subjects=np.expand_dims(self.data_subjects, axis=axis),
        )

    def ones_like(
        self,
        *args: Tuple[Any, ...],
        **kwargs: Any,
    ) -> Union[PhiTensor, GammaTensor]:
        # TODO: Add support for axes arguments later
        min_vals = self.min_vals.ones_like(*args, **kwargs)
        max_vals = self.max_vals.ones_like(*args, **kwargs)

        child = (
            np.ones_like(self.child, *args, **kwargs)
            if isinstance(self.child, np.ndarray)
            else self.child.ones_like(*args, **kwargs)
        )

        return PhiTensor(
            child=child,
            min_vals=min_vals,
            max_vals=max_vals,
            data_subjects=self.data_subjects,
        )

    def __ne__(self, other: Any) -> Union[PhiTensor, GammaTensor]:  # type: ignore
        # TODO: what about data_subjects and min / max values?
        if is_acceptable_simple_type(other) or len(self.child) == len(other.child):
            gamma_output = False
            if is_acceptable_simple_type(other):
                result = self.child != other
            else:
                # check data_subjects match, if they dont gamma_output = True
                #
                result = self.child != other.child
                if isinstance(result, GammaTensor):
                    gamma_output = True
            if not gamma_output:
                return self.copy_with(child=result)
            else:
                return self.copy_with(child=result).gamma
        else:
            raise Exception(
                "Tensor dims do not match for __eq__: "
                + f"{len(self.child)} != {len(other.child)}"
            )

    def __neg__(self) -> PhiTensor:

        return PhiTensor(
            child=self.child * -1,
            min_vals=self.max_vals * -1,
            max_vals=self.min_vals * -1,
            data_subjects=self.data_subjects,
        )

    def __pos__(self) -> PhiTensor:
        return PhiTensor(
            child=self.child,
            min_vals=self.min_vals,
            max_vals=self.max_vals,
            data_subjects=self.data_subjects,
        )

    def exp(self) -> PhiTensor:
        # relative
        from ...smpc.approximations import exp

        def exp_reduction(val: np.ndarray) -> np.ndarray:
            pos_index = val >= 0
            neg_index = val < 0
            exp = np.exp((pos_index * val * -1) + (neg_index * val))
            pos_values = (pos_index) * exp
            neg_values = (neg_index) * exp * -1
            return pos_values + neg_values

        min_vals = self.min_vals.copy()
        min_vals.data = np.array(exp_reduction(min_vals.data))
        max_vals = self.max_vals.copy()
        max_vals.data = np.array(exp_reduction(max_vals.data))

        return PhiTensor(
            child=exp(self.child),  # type: ignore
            min_vals=min_vals,
            max_vals=max_vals,
            data_subjects=self.data_subjects,
        )

    def softmax(self) -> PhiTensor:
        # relative
        from ...smpc.approximations import exp
        from ...smpc.approximations import reciprocal

        def softmax(val: np.ndarray) -> np.ndarray:
            logits = val - val.max()
            numerator = np.exp(logits)
            inv = 1 / numerator.sum()
            return numerator * inv

        min_vals = self.min_vals.copy()
        min_vals.data = np.array(softmax(min_vals.data))
        max_vals = self.max_vals.copy()
        max_vals.data = np.array(softmax(max_vals.data))
        fpt = self.child.copy()
        if not isinstance(fpt.child, np.ndarray):
            raise ValueError("Softmax currently works only for numpy child")

        fpt.child = fpt.child - fpt.child.max()
        numerator = exp(fpt)
        inv = reciprocal(numerator.sum())  # type: ignore

        return PhiTensor(
            child=numerator * inv,  # type: ignore
            min_vals=min_vals,
            max_vals=max_vals,
            data_subjects=self.data_subjects,
        )

    def reciprocal(self) -> PhiTensor:
        # relative
        from ...smpc.approximations import reciprocal

        min_vals = self.min_vals.copy()
        min_vals.data = np.array(1 / min_vals.data)
        max_vals = self.max_vals.copy()
        max_vals.data = np.array(1 / max_vals.data)

        return PhiTensor(
            child=reciprocal(self.child),
            min_vals=min_vals,
            max_vals=max_vals,
            data_subjects=self.data_subjects,
        )

    def one_hot(self) -> PhiTensor:
        one_hot_child = self.child.one_hot()

        return PhiTensor(
            child=one_hot_child,
            min_vals=self.min_vals,
            max_vals=self.max_vals,
            data_subjects=self.data_subjects,
        )

    def _object2bytes(self) -> bytes:
        schema = get_capnp_schema(schema_file="phi_tensor.capnp")

        pt_struct: CapnpModule = schema.PT  # type: ignore
        pt_msg = pt_struct.new_message()
        # this is how we dispatch correct deserialization of bytes
        pt_msg.magicHeader = serde_magic_header(type(self))

        if isinstance(self.child, np.ndarray) or np.isscalar(self.child):
            chunk_bytes(capnp_serialize(np.array(self.child), to_bytes=True), "child", pt_msg)  # type: ignore
            pt_msg.isNumpy = True
        else:
            chunk_bytes(serialize(self.child, to_bytes=True), "child", pt_msg)  # type: ignore
            pt_msg.isNumpy = False

        pt_msg.minVals = serialize(self.min_vals, to_bytes=True)
        pt_msg.maxVals = serialize(self.max_vals, to_bytes=True)
        chunk_bytes(
            capnp_serialize(dslarraytonumpyutf8(self.data_subjects), to_bytes=True),
            "dataSubjects",
            pt_msg,
        )
        # to pack or not to pack?
        # to_bytes = pt_msg.to_bytes()

        return pt_msg.to_bytes_packed()

    @staticmethod
    def _bytes2object(buf: bytes) -> PhiTensor:
        schema = get_capnp_schema(schema_file="phi_tensor.capnp")
        pt_struct: CapnpModule = schema.PT  # type: ignore
        # https://stackoverflow.com/questions/48458839/capnproto-maximum-filesize
        MAX_TRAVERSAL_LIMIT = 2**64 - 1
        # to pack or not to pack?
        # pt_msg = pt_struct.from_bytes(buf, traversal_limit_in_words=2 ** 64 - 1)
        pt_msg = pt_struct.from_bytes_packed(
            buf, traversal_limit_in_words=MAX_TRAVERSAL_LIMIT
        )

        if pt_msg.isNumpy:
            child = capnp_deserialize(combine_bytes(pt_msg.child), from_bytes=True)
        else:
            child = deserialize(combine_bytes(pt_msg.child), from_bytes=True)

        min_vals = deserialize(pt_msg.minVals, from_bytes=True)
        max_vals = deserialize(pt_msg.maxVals, from_bytes=True)
        data_subjects = numpyutf8todslarray(
            capnp_deserialize(combine_bytes(pt_msg.dataSubjects), from_bytes=True)
        )

        return PhiTensor(
            child=child,
            min_vals=min_vals,
            max_vals=max_vals,
            data_subjects=data_subjects,
        )



# stdlib
from typing import Any
from typing import List

# third party
from sqlalchemy.engine import Engine
from sqlalchemy.orm import sessionmaker

# relative
from ..node_table.pdf import PDFObject
from ..node_table.setup import SetupConfig

# from ..exceptions import SetupNotFoundError
from .database_manager import DatabaseManager


class SetupManager(DatabaseManager):

    schema = SetupConfig

    def __init__(self, database: Engine) -> None:
        super().__init__(db=database, schema=SetupManager.schema)

    def register_once(self, **kwargs: Any) -> Any:
        """Register a new object into the database.

        Args:
            parameters : List of object parameters.
        Returns:
            object: Database Object
        """
        _obj = self._schema(**kwargs)
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()

        with session_local.begin():
            nrows = len(self)
            if nrows == 0:
                session_local.add(_obj)
                session_local.commit()

        return _obj

    @property
    def node_name(self) -> str:
        setup = super().all()[0]
        return setup.domain_name

    @property
    def id(self) -> int:
        setup = super().all()[0]
        return setup.id

    def first(self, **kwargs: Any) -> SetupConfig:
        result = super().first(**kwargs)
        if not result:
            # raise SetupNotFoundError
            raise Exception
        return result

    def query(self, **kwargs: Any) -> List[SetupConfig]:
        results = super().query(**kwargs)
        if len(results) == 0:
            # raise SetupNotFoundError
            raise Exception
        return results

    def update(self, **kwargs: Any) -> None:
        session_local = sessionmaker(autocommit=False, autoflush=False, bind=self.db)()
        if "daa_document" in kwargs.keys():
            _pdf_obj = PDFObject(binary=kwargs["daa_document"])
            session_local.add(_pdf_obj)
            session_local.commit()
            session_local.flush()
            session_local.refresh(_pdf_obj)
            kwargs["daa_document"] = str(_pdf_obj.id)

        session_local.query(self._schema).update(kwargs)
        session_local.commit()
        session_local.close()



# future
from __future__ import annotations

# stdlib
import logging
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Type

# third party
from pandas import DataFrame

# relative
from .....experimental_flags import flags
from ....common.message import SyftMessage  # type: ignore
from ...abstract.node import AbstractNodeClient
from ...enums import RequestAPIFields
from ..action.exception_action import ExceptionMessage
from ..node_service.generic_payload.messages import GenericPayloadMessageWithReply
from ..node_service.generic_payload.syft_message import NewSyftMessage


class RequestAPI:
    def __init__(
        self,
        client: AbstractNodeClient,
        create_msg: Optional[Type[SyftMessage] | Type[NewSyftMessage]] = None,
        get_msg: Optional[Type[SyftMessage] | Type[NewSyftMessage]] = None,
        get_all_msg: Optional[Type[SyftMessage] | Type[NewSyftMessage]] = None,
        update_msg: Optional[Type[SyftMessage] | Type[NewSyftMessage]] = None,
        delete_msg: Optional[Type[SyftMessage] | Type[NewSyftMessage]] = None,
        response_key: str = "",
    ):
        self._create_message = create_msg
        self._get_message = get_msg
        self._get_all_message = get_all_msg
        self._update_message = update_msg
        self._delete_message = delete_msg
        self._response_key = response_key
        self.client = client
        self.perform_request = self.perform_api_request_generic

    def create(self, **kwargs: Any) -> None:
        if self._create_message and issubclass(self._create_message, NewSyftMessage):
            response = self.perform_request(  # type: ignore
                syft_msg=self._create_message, content=kwargs
            )
            logging.info(response.message)
        else:
            timeout = kwargs.pop("timeout")
            response = self.perform_api_request(  # type: ignore
                syft_msg=self._create_message, content=kwargs, timeout=timeout
            )
            logging.info(response.resp_msg)

    def get(self, **kwargs: Any) -> Any:
        if isinstance(self._get_message, NewSyftMessage):
            return self.to_obj(  # type: ignore
                self.perform_request(syft_msg=self._get_message, content=kwargs)
            )
        else:
            return self.to_obj(
                self.perform_api_request(syft_msg=self._get_message, content=kwargs)
            )

    def all(self) -> List[Any]:
        if self._get_all_message and issubclass(self._get_all_message, NewSyftMessage):
            result = self.perform_request(syft_msg=self._get_all_message).kwargs[
                "users"
            ]
        else:
            result = []
            for content in self.perform_api_request(
                syft_msg=self._get_all_message
            ).content:
                if hasattr(content, "upcast"):
                    content = content.upcast()
                result.append(content)
        return result

    def pandas(self) -> DataFrame:
        return DataFrame(self.all())

    def update(self, **kwargs: Any) -> None:
        if isinstance(self._delete_message, NewSyftMessage):
            response = self.perform_request(  # type: ignore
                syft_msg=self._update_message, content=kwargs
            )
            logging.info(response.message)
        else:
            response = self.perform_api_request(
                syft_msg=self._delete_message, content=kwargs
            )
            logging.info(response.resp_msg)

    def delete(self, **kwargs: Any) -> None:
        if isinstance(self._delete_message, NewSyftMessage):
            response = self.perform_request(  # type: ignore
                syft_msg=self._delete_message, content=kwargs
            )
            logging.info(response.message)
        else:
            response = self.perform_api_request(
                syft_msg=self._delete_message, content=kwargs
            )

    def to_obj(self, result: Any) -> Any:
        if result:
            _class_name = self._response_key.capitalize()
            if flags.USE_NEW_SERVICE:
                result = type(_class_name, (object,), result.dict())()
            else:
                result = type(_class_name, (object,), result)()

        return result

    def perform_api_request(
        self,
        syft_msg: Optional[Type[SyftMessage]],
        content: Optional[Dict[Any, Any]] = None,
        timeout: Optional[int] = None,
    ) -> Any:
        if syft_msg is None:
            raise ValueError(
                "Can't perform this type of api request, the message is None."
            )
        else:
            syft_msg_constructor = syft_msg

        if content is None:
            content = {}
        content[RequestAPIFields.ADDRESS] = self.client.address
        content[RequestAPIFields.REPLY_TO] = self.client.address

        signed_msg = syft_msg_constructor(**content).sign(
            signing_key=self.client.signing_key
        )  # type: ignore
        response = self.client.send_immediate_msg_with_reply(
            msg=signed_msg, timeout=timeout
        )
        if isinstance(response, ExceptionMessage):
            raise response.exception_type
        else:
            return response

    def perform_api_request_generic(
        self,
        syft_msg: Optional[Type[GenericPayloadMessageWithReply] | Type[NewSyftMessage]],  # type: ignore
        content: Optional[Dict[Any, Any]] = None,
        timeout: Optional[int] = None,
    ) -> Any:
        if syft_msg is None:
            raise ValueError(
                "Can't perform this type of api request, the message is None."
            )
        else:
            syft_msg_constructor = syft_msg

        if content is None:
            content = {}

        if issubclass(syft_msg, NewSyftMessage):
            signed_msg = syft_msg_constructor(  # type: ignore
                address=self.client.address, reply_to=self.client.address, kwargs=content  # type: ignore
            ).sign(  # type: ignore
                signing_key=self.client.signing_key
            )
        else:
            signed_msg = (
                syft_msg_constructor(kwargs=content)  # type: ignore
                .to(  # type: ignore
                    address=self.client.address, reply_to=self.client.address
                )
                .sign(signing_key=self.client.signing_key)
            )
        response = self.client.send_immediate_msg_with_reply(
            msg=signed_msg, timeout=timeout
        )
        if isinstance(response, ExceptionMessage):
            raise response.exception_type
        else:
            if isinstance(syft_msg, NewSyftMessage):
                return response.payload  # type: ignore
            return response

    def _repr_html_(self) -> str:
        return self.pandas()._repr_html_()



# stdlib
from typing import Any

# third party
from sqlalchemy import Column
from sqlalchemy import LargeBinary
from sqlalchemy import String

# relative
from . import Base
from ..... import deserialize
from ..... import serialize


class Entity(Base):
    __tablename__ = "entity"

    name = Column(String(255), primary_key=True)
    entity_bin = Column(LargeBinary(3072))

    @property
    def obj(self) -> Any:
        return deserialize(self.entity_bin, from_bytes=True)  # TODO: techdebt fix

    @obj.setter
    def obj(self, value: Any) -> None:
        self.entity_bin = serialize(value, to_bytes=True)  # TODO: techdebt fix



"""SPDZ Protocol.

SPDZ mechanism used for multiplication Contains functions that are run at:

* the party that orchestrates the computation
* the parties that hold the shares
"""

# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Dict
from typing import TYPE_CHECKING

# relative
from ....tensor.config import DEFAULT_RING_SIZE
from ....tensor.smpc import utils
from ...store import CryptoPrimitiveProvider

EXPECTED_OPS = {"mul", "matmul"}

if TYPE_CHECKING:
    # relative
    from ....tensor.smpc.mpc_tensor import MPCTensor


def mul_master(
    x: MPCTensor, y: MPCTensor, op_str: str, **kwargs: Dict[Any, Any]
) -> MPCTensor:

    """Function that is executed by the orchestrator to multiply two secret values.

    Args:
        x (MPCTensor): First value to multiply with.
        y (MPCTensor): Second value to multiply with.
        op_str (str): Operation string.

    Raises:
        ValueError: If op_str not in EXPECTED_OPS.

    Returns:
        MPCTensor: Result of the multiplication.
    """

    parties = x.parties
    parties_info = x.parties_info

    ring_size = utils.get_ring_size(x.ring_size, y.ring_size)
    shape_x = tuple(x.shape)  # type: ignore
    shape_y = tuple(y.shape)  # type: ignore
    result_shape = utils.get_shape(op_str, shape_x, shape_y)
    if ring_size != 2:
        # For ring_size 2 we generate those before hand
        CryptoPrimitiveProvider.generate_primitives(
            f"beaver_{op_str}",
            parties=parties,
            g_kwargs={
                "a_shape": shape_x,
                "b_shape": shape_y,
                "parties_info": parties_info,
            },
            p_kwargs={"a_shape": shape_x, "b_shape": shape_y},
            ring_size=ring_size,
        )
        # TODO: Should input size after the computation of a dummy function
        # This will not work for matmul
        CryptoPrimitiveProvider.generate_primitives(
            "beaver_wraps",
            parties=parties,
            g_kwargs={
                "shape": result_shape,
                "parties_info": parties_info,
            },
            p_kwargs={"shape": result_shape},
            ring_size=ring_size,
        )

    # TODO: Should modify to parallel execution.

    res_shares = [
        getattr(a, f"__{op_str}__")(b, **kwargs) for a, b in zip(x.child, y.child)
    ]

    return res_shares  # type: ignore


def lt_master(x: MPCTensor, y: MPCTensor, op_str: str) -> MPCTensor:
    """Function that is executed by the orchestrator to multiply two secret values.

    Args:
        x (MPCTensor): First value to multiply with.
        y (MPCTensor): Second value to multiply with.
        op_str (str): Operation string.

    Raises:
        ValueError: If op_str not in EXPECTED_OPS.

    Returns:
        MPCTensor: Result of the multiplication.
    """
    # relative
    # from ....tensor.smpc.mpc_tensor import MPCTensor
    # from ....tensor.tensor import TensorPointer

    # diff = a - b
    # bit decomposition
    # sum carry adder
    # res = sign(diff)

    res_shares = x - y
    res_shares.block
    # time.sleep(2)
    msb = MSB(res_shares)

    # This solves the high budget spent in DP operations,
    # This code is to be removed when we move comparison to ShareTensor level.
    # tensor_shares = []
    # final_shares = []

    # if isinstance(x, MPCTensor):
    #     if isinstance(x.child[0], TensorPointer):
    #         for t1, t2 in zip(x.child, y.child):
    #             tensor_shares.append(t1.__lt__(t2))

    #         for p1, p2 in zip(tensor_shares, msb.child):
    #             p2.block
    #             final_shares.append(p1.mpc_swap(p2))

    #         msb.child = final_shares
    # else:
    #     if isinstance(y.child[0], TensorPointer):  # type: ignore
    #         for t1 in y.child:
    #             tensor_shares.append(t1.__lt__(x))

    #         for p1, p2 in zip(tensor_shares, msb.child):
    #             p2.block
    #             final_shares.append(p1.mpc_swap(p2))

    #         msb.child = final_shares

    return msb


def MSB(x: MPCTensor) -> MPCTensor:
    # relative
    from ..aby3 import ABY3

    """Computes the MSB of the underlying share

    Args:
        x (MPCTensor): Input MPCTensor to compute MSB on.

    Returns:
        msb (MPCTensor): returns arithmetic shares of the MSB.
    """
    ring_size = DEFAULT_RING_SIZE
    decomposed_shares = ABY3.bit_decomposition(x)
    # return decomposed_shares

    for share in decomposed_shares:
        share.block

    msb_share = decomposed_shares[-1]
    msb_share.block

    msb = ABY3.bit_injection(msb_share, ring_size)

    return msb



# stdlib
import re
from typing import Any
from typing import Dict
from typing import List

# third party
from packaging import version
import torch

# relative
from ...core.common.serde.serializable import serializable
from ...proto.lib.torch.returntypes_pb2 import ReturnTypes as ReturnTypes_PB
from ..util import full_name_with_name
from .tensor_util import tensor_deserializer
from .tensor_util import tensor_serializer

# TODO: a better way. Loot at https://github.com/OpenMined/PySyft/issues/5249
module_type = type(torch)
torch.__dict__["return_types"] = module_type(name="return_types")
parent = torch.__dict__["return_types"]


def get_field_names(obj: Any) -> List[str]:
    return re.findall("\n(.*)=", str(obj))


def get_supported_types_fields() -> Dict[type, List]:
    supported_types = {}
    # A = torch.tensor([[1.0, 1, 1], [2, 3, 4], [3, 5, 2], [4, 2, 5], [5, 4, 3]])
    # B = torch.tensor([[-10.0, -3], [12, 14], [14, 12], [16, 16], [18, 16]])
    x = torch.Tensor([[1, 2], [1, 2]])
    s = torch.tensor(
        [[-0.1000, 0.1000, 0.2000], [0.2000, 0.3000, 0.4000], [0.0000, -0.3000, 0.5000]]
    )

    torch_version_ge_1d5d0 = version.parse(
        torch.__version__.split("+")[0]
    ) >= version.parse("1.5.0")

    if torch_version_ge_1d5d0:
        cummax = x.cummax(0)
        supported_types[type(cummax)] = get_field_names(cummax)

    if torch_version_ge_1d5d0:
        cummin = x.cummin(0)
        supported_types[type(cummin)] = get_field_names(cummin)

    # deprecated in torch==1.10.0
    # eig = x.eig(True)
    # supported_types[type(eig)] = get_field_names(eig)

    kthvalue = x.kthvalue(1)
    supported_types[type(kthvalue)] = get_field_names(kthvalue)

    # deprecated in torch==1.10.0
    # lstsq = A.lstsq(B)
    # supported_types[type(lstsq)] = get_field_names(lstsq)

    slogdet = x.slogdet()
    supported_types[type(slogdet)] = get_field_names(slogdet)

    # deprecated in torch==1.10.0
    # qr = x.qr()
    # supported_types[type(qr)] = get_field_names(qr)

    mode = x.mode()
    supported_types[type(mode)] = get_field_names(mode)

    # deprecated in torch==1.10.0
    # solve = s.solve(s)
    # supported_types[type(solve)] = get_field_names(solve)

    sort = s.sort()
    supported_types[type(sort)] = get_field_names(sort)

    # deprecated in torch==1.10.0
    # symeig = s.symeig()
    # supported_types[type(symeig)] = get_field_names(symeig)

    topk = s.topk(1)
    supported_types[type(topk)] = get_field_names(topk)

    # deprecated in torch==1.11.0
    # triangular_solve = s.triangular_solve(s)
    # supported_types[type(triangular_solve)] = get_field_names(triangular_solve)

    svd = s.svd()
    supported_types[type(svd)] = get_field_names(svd)

    geqrf = s.geqrf()
    supported_types[type(geqrf)] = get_field_names(geqrf)

    median = s.median(0)
    supported_types[type(median)] = get_field_names(median)

    max_t = s.max(0)
    supported_types[type(max_t)] = get_field_names(max_t)

    min_t = s.min(0)
    supported_types[type(min_t)] = get_field_names(min_t)

    return supported_types


def wrap_type(typ: type, fields: List[str]) -> None:
    def object2proto(obj: object) -> ReturnTypes_PB:
        proto = ReturnTypes_PB()

        obj_type = full_name_with_name(klass=obj._sy_serializable_wrapper_type)  # type: ignore
        proto.obj_type = obj_type

        values = [getattr(obj, field, None) for field in fields]
        proto.values.extend(list(map(lambda x: tensor_serializer(x), values)))

        return proto

    def proto2object(proto: ReturnTypes_PB) -> "typ":  # type: ignore
        values = [tensor_deserializer(x) for x in proto.values]
        return typ(values)

    serializable(generate_wrapper=True)(
        wrapped_type=typ,
        import_path=f"{typ.__module__}.{typ.__name__}",
        protobuf_scheme=ReturnTypes_PB,
        type_object2proto=object2proto,
        type_proto2object=proto2object,
    )

    # TODO: a better way. Loot at https://github.com/OpenMined/PySyft/issues/5249
    # add type to torch.return_types
    parent.__dict__[typ.__name__] = typ


types_fields = get_supported_types_fields()
for typ, fields in types_fields.items():
    wrap_type(typ, fields)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: proto/grid/messages/infra_messages.proto
"""Generated protocol buffer code."""
# third party
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


# syft absolute
from syft.proto.core.common import (
    common_object_pb2 as proto_dot_core_dot_common_dot_common__object__pb2,
)
from syft.proto.core.io import address_pb2 as proto_dot_core_dot_io_dot_address__pb2

DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n(proto/grid/messages/infra_messages.proto\x12\x12syft.grid.messages\x1a%proto/core/common/common_object.proto\x1a\x1bproto/core/io/address.proto"\x9e\x01\n\x13\x43reateWorkerMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8b\x01\n\x14\x43reateWorkerResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\xa8\x01\n\x1dGetWorkerInstanceTypesMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x95\x01\n\x1eGetWorkerInstanceTypesResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x9b\x01\n\x10GetWorkerMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x88\x01\n\x11GetWorkerResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x9c\x01\n\x11GetWorkersMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x89\x01\n\x12GetWorkersResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x9e\x01\n\x13\x44\x65leteWorkerMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8b\x01\n\x14\x44\x65leteWorkerResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x9e\x01\n\x13UpdateWorkerMessage\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12&\n\x07\x61\x64\x64ress\x18\x02 \x01(\x0b\x32\x15.syft.core.io.Address\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12\'\n\x08reply_to\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Address"\x8b\x01\n\x14UpdateWorkerResponse\x12%\n\x06msg_id\x18\x01 \x01(\x0b\x32\x15.syft.core.common.UID\x12\x13\n\x0bstatus_code\x18\x02 \x01(\x05\x12\x0f\n\x07\x63ontent\x18\x03 \x01(\t\x12&\n\x07\x61\x64\x64ress\x18\x04 \x01(\x0b\x32\x15.syft.core.io.Addressb\x06proto3'
)

_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, globals())
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "proto.grid.messages.infra_messages_pb2", globals()
)
if _descriptor._USE_C_DESCRIPTORS == False:

    DESCRIPTOR._options = None
    _CREATEWORKERMESSAGE._serialized_start = 133
    _CREATEWORKERMESSAGE._serialized_end = 291
    _CREATEWORKERRESPONSE._serialized_start = 294
    _CREATEWORKERRESPONSE._serialized_end = 433
    _GETWORKERINSTANCETYPESMESSAGE._serialized_start = 436
    _GETWORKERINSTANCETYPESMESSAGE._serialized_end = 604
    _GETWORKERINSTANCETYPESRESPONSE._serialized_start = 607
    _GETWORKERINSTANCETYPESRESPONSE._serialized_end = 756
    _GETWORKERMESSAGE._serialized_start = 759
    _GETWORKERMESSAGE._serialized_end = 914
    _GETWORKERRESPONSE._serialized_start = 917
    _GETWORKERRESPONSE._serialized_end = 1053
    _GETWORKERSMESSAGE._serialized_start = 1056
    _GETWORKERSMESSAGE._serialized_end = 1212
    _GETWORKERSRESPONSE._serialized_start = 1215
    _GETWORKERSRESPONSE._serialized_end = 1352
    _DELETEWORKERMESSAGE._serialized_start = 1355
    _DELETEWORKERMESSAGE._serialized_end = 1513
    _DELETEWORKERRESPONSE._serialized_start = 1516
    _DELETEWORKERRESPONSE._serialized_end = 1655
    _UPDATEWORKERMESSAGE._serialized_start = 1658
    _UPDATEWORKERMESSAGE._serialized_end = 1816
    _UPDATEWORKERRESPONSE._serialized_start = 1819
    _UPDATEWORKERRESPONSE._serialized_end = 1958
# @@protoc_insertion_point(module_scope)



# stdlib
from typing import Any
from typing import Optional
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType

# syft absolute
import syft as sy

# relative
from ...core.common import UID
from ...core.common.serde.serializable import serializable
from ...proto.lib.python.range_pb2 import Range as Range_PB
from .iterator import Iterator
from .primitive_factory import PrimitiveFactory
from .primitive_interface import PyPrimitive
from .types import SyPrimitiveRet


@serializable()
class Range(PyPrimitive):
    __slots__ = ["_id", "_index"]

    def __init__(
        self,
        start: Any = None,
        stop: Union[Any] = None,
        step: Union[Any] = 1,
        id: Optional[UID] = None,
    ):
        if stop is None:
            stop = start
            start = 0
        self.value = range(start, stop, step)
        self._id: UID = id if id else UID()

    @property
    def id(self) -> UID:
        """We reveal PyPrimitive.id as a property to discourage users and
        developers of Syft from modifying .id attributes after an object
        has been initialized.

        :return: returns the unique id of the object
        :rtype: UID
        """
        return self._id

    def __contains__(self, other: Any) -> SyPrimitiveRet:
        res = self.value.__contains__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __eq__(self, other: Any) -> SyPrimitiveRet:
        res = self.value.__eq__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __ne__(self, other: Any) -> SyPrimitiveRet:
        res = self.value.__ne__(other)
        return PrimitiveFactory.generate_primitive(value=res)

    def __sizeof__(self) -> SyPrimitiveRet:
        res = self.value.__sizeof__()
        return PrimitiveFactory.generate_primitive(value=res)

    def __bool__(self) -> SyPrimitiveRet:
        # res = self.value.__bool__()
        # mypy error: "range" has no attribute "__bool__"
        # work around:
        try:
            res = bool(self.value.__len__())
        except OverflowError:
            res = True
        return PrimitiveFactory.generate_primitive(value=res)

    def __len__(self) -> Any:
        res = self.value.__len__()
        return PrimitiveFactory.generate_primitive(value=res)

    def __getitem__(self, key: Union[int]) -> Any:
        res = self.value.__getitem__(key)
        return PrimitiveFactory.generate_primitive(value=res)

    def __iter__(self, max_len: Optional[int] = None) -> Iterator:
        return Iterator(self.value, max_len=max_len)

    @property
    def start(self) -> SyPrimitiveRet:
        res = self.value.start
        return PrimitiveFactory.generate_primitive(value=res)

    @property
    def step(self) -> SyPrimitiveRet:
        res = self.value.step
        return PrimitiveFactory.generate_primitive(value=res)

    @property
    def stop(self) -> SyPrimitiveRet:
        res = self.value.stop
        return PrimitiveFactory.generate_primitive(value=res)

    def index(self, value: int) -> SyPrimitiveRet:
        res = self.value.index(value)
        return PrimitiveFactory.generate_primitive(value=res)

    def count(self, value: int) -> SyPrimitiveRet:
        res = self.value.count(value)
        return PrimitiveFactory.generate_primitive(value=res)

    def upcast(self) -> range:
        return self.value

    def _object2proto(self) -> Range_PB:
        range_pb = Range_PB()

        range_pb.start = self.start
        range_pb.stop = self.stop
        range_pb.step = self.step
        range_pb.id.CopyFrom(self._id._object2proto())

        return range_pb

    @staticmethod
    def _proto2object(proto: Range_PB) -> "Range":

        return Range(
            start=proto.start,
            stop=proto.stop,
            step=proto.step,
            id=sy.deserialize(blob=proto.id),
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        return Range_PB



# stdlib
from typing import Any
from typing import Optional
from typing import Tuple
from typing import Union

# third party
import numpy as np
from numpy.typing import NDArray

# relative
from .. import activations
from ....common.serde.serializable import serializable
from ...autodp.gamma_tensor import GammaTensor
from ...autodp.phi_tensor import PhiTensor
from ..initializations import XavierInitialization
from ..utils import col2im_indices
from ..utils import im2col_indices
from .base import Layer


@serializable(recursive_serde=True)
class Convolution(Layer):
    """
    If this is the first layer in a model, provide the keyword argument `input_shape`
    (tuple of integers, does NOT include the sample axis, N.),
    e.g. `input_shape=(3, 128, 128)` for 128x128 RGB pictures.
    """

    __attr_allowlist__ = [
        "nb_filter",
        "filter_size",
        "input_shape",
        "stride",
        "padding",
        "W",
        "b",
        "dW",
        "db",
        "out_shape",
        "last_output",
        "last_input",
        "init",
        "activation",
    ]

    def __init__(
        self,
        nb_filter: int,
        filter_size: Union[int, Tuple[int, int]],
        input_shape: Optional[Tuple] = None,
        stride: int = 1,
        padding: int = 0,
        activation: Optional[str] = None,
    ):
        self.nb_filter = nb_filter
        self.filter_size = filter_size
        self.input_shape = input_shape
        self.stride = stride
        self.padding = padding
        self.name = "Conv"

        self.W: Optional[Union[NDArray, PhiTensor, GammaTensor]] = None
        self.dW: Optional[Union[PhiTensor, GammaTensor]] = None
        self.b: Optional[Union[NDArray, PhiTensor, GammaTensor]] = None
        self.db: Optional[Union[PhiTensor, GammaTensor]] = None
        self.out_shape: Optional[Tuple[int, ...]] = None
        self.last_output: Optional[Union[PhiTensor, GammaTensor]] = None
        self.last_input: Optional[Union[PhiTensor, GammaTensor]] = None

        self.init = XavierInitialization()
        self.activation = activations.get(activation)

    def connect_to(self, prev_layer: Optional[Layer] = None) -> None:
        if prev_layer is None:
            if self.input_shape is None:
                raise ValueError(
                    "`input_shape` cannot be None. \
                    Please specify the input shape, since this is the first layer."
                )
            input_shape = self.input_shape
        else:
            input_shape = prev_layer.out_shape  # type: ignore

        if len(input_shape) != 4:
            raise ValueError(
                "Input shape to the layer should be of the format: (N, C, H, W)."
            )

        nb_batch, pre_nb_filter, pre_height, pre_width = input_shape
        if isinstance(self.filter_size, tuple):
            filter_height, filter_width = self.filter_size
        elif isinstance(self.filter_size, int):
            filter_height = filter_width = self.filter_size
        else:
            raise TypeError(
                "Argument `filter_size` should either to be Tuple[int, int] or an integer."
            )

        height = (pre_height - filter_height + 2 * self.padding) // self.stride + 1
        width = (pre_width - filter_width + 2 * self.padding) // self.stride + 1

        # output shape
        self.out_shape = (nb_batch, self.nb_filter, height, width)

        # filters
        self.W = self.init((self.nb_filter, pre_nb_filter, filter_height, filter_width))
        self.b = np.zeros((self.nb_filter,))

    def forward(
        self,
        input: Union[PhiTensor, GammaTensor],
        *args: Optional[Any],
        **kwargs: Optional[Any],
    ) -> Union[PhiTensor, GammaTensor]:

        self.last_input = input
        self.input_shape = input.shape

        if self.W is None or self.b is None:
            raise ValueError(
                """Weight or bias for the layer cannot be None.
                Please initialize the weight (self.W) and bias (self.b) for the layer."""
            )

        # N, C, H, W
        n_filters, _, h_filter, w_filter = self.W.shape

        # Input Shape -> N, C, H, W
        n_x, _, _, _ = input.shape

        if self.out_shape is None:
            raise ValueError(
                "`self.out_shape` cannot be None. Please check if the output_shape to the layer is correctly defined."
            )

        _, _, h_out, w_out = self.out_shape

        self.X_col: Union[GammaTensor, PhiTensor] = im2col_indices(
            input, h_filter, w_filter, padding=self.padding, stride=self.stride
        )

        W_col = self.W.reshape((n_filters, -1))
        out = (
            self.X_col.T @ W_col.T + self.b
        )  # Transpose is required here because W_col is numpy array
        out = out.reshape((n_filters, h_out, w_out, n_x))
        out = out.transpose((3, 0, 1, 2))

        self.last_output = (
            self.activation.forward(out) if self.activation is not None else out
        )

        return out

    def backward(
        self,
        pre_grad: Union[PhiTensor, GammaTensor],
        *args: Optional[Any],
        **kwargs: Optional[Any],
    ) -> Union[PhiTensor, GammaTensor]:

        if self.W is None or self.b is None:
            raise ValueError(
                """Weight or bias for the layer cannot be None.
                Please initialize the weight (self.W) and bias (self.b) for the layer."""
            )

        # N, C, H, W
        n_filter, _, h_filter, w_filter = self.W.shape

        pre_grads = (
            self.activation.derivative(pre_grad)
            if self.activation is not None
            else pre_grad
        )

        db = pre_grads.sum(axis=(0, 2, 3))

        self.db = db.reshape((n_filter,))

        pre_grads_reshaped = pre_grads.transpose((1, 2, 3, 0))
        pre_grads_reshaped = pre_grads_reshaped.reshape((n_filter, -1))
        pre_grads_reshaped.min_vals.shape = (
            pre_grads_reshaped.max_vals.shape
        ) = pre_grads_reshaped.shape
        dW = pre_grads_reshaped @ self.X_col.T
        self.dW = dW.reshape((self.W.shape))

        W_reshape = self.W.reshape((n_filter, -1))
        dX_col = pre_grads_reshaped.T @ W_reshape

        if self.input_shape is None:
            raise ValueError(
                "Input shape to the layer cannot be None. \
                Please check if the input shape has been initialized ."
            )
        dX = col2im_indices(
            dX_col,
            self.input_shape,
            h_filter,
            w_filter,
            padding=self.padding,
            stride=self.stride,
        )
        return dX

    @property
    def params(self) -> Tuple[Union[PhiTensor, GammaTensor, NDArray, None], ...]:
        return self.W, self.b

    @params.setter
    def params(
        self, new_params: Tuple[Union[PhiTensor, GammaTensor, NDArray], ...]
    ) -> None:

        if len(new_params) != 2:
            raise ValueError(
                f"Expected two values. Update params has length {len(new_params)}"
            )
        self.W, self.b = new_params

    @property
    def grads(self) -> Tuple[Union[PhiTensor, GammaTensor, NDArray, None], ...]:
        return self.dW, self.db



# stdlib
from typing import Any
from typing import Dict

# third party
from sqlalchemy import inspect
from sqlalchemy.ext.declarative import as_declarative


@as_declarative()
class Base:
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        pass

    id: Any
    __name__: str

    def _asdict(self) -> Dict:
        return {c.key: getattr(self, c.key) for c in inspect(self).mapper.column_attrs}



# stdlib
from typing import Dict
from typing import Optional
from typing import Tuple
from typing import Type
from typing import Union

# relative
from ...common.serde.serializable import serializable
from ..autodp.gamma_tensor import GammaTensor
from ..autodp.phi_tensor import PhiTensor


@serializable(recursive_serde=True)
class Activation(object):
    """Base class for activations."""

    __attr_allowlist__: Tuple[str, ...] = ("last_forward",)

    def __init__(self) -> None:
        self.last_forward: Optional[Union[PhiTensor, GammaTensor]] = None

    def forward(
        self, input: Union[PhiTensor, GammaTensor]
    ) -> Union[PhiTensor, GammaTensor]:
        """Forward Step.

        Args:
            input (PhiTensor or GammaTensor): the input matrix
        """

        raise NotImplementedError

    def derivative(
        self, input: Optional[Union[PhiTensor, GammaTensor]] = None
    ) -> Union[PhiTensor, GammaTensor]:
        """Backward Step.

        _extended_summary_

        Args:
            input (Optional[PhiTensor, GammaTensor], optional): If provide `input`,
            this function will not use `last_forward`. Defaults to None.
        """

        raise NotImplementedError

    def __str__(self) -> str:
        return self.__class__.__name__


@serializable(recursive_serde=True)
class leaky_ReLU(Activation):
    __attr_allowlist__ = (
        "slope",
        "last_forward",
    )

    def __init__(self, slope: float = 0.01) -> None:
        super(leaky_ReLU, self).__init__()
        self.slope = slope

    def forward(
        self, input_array: Union[PhiTensor, GammaTensor]
    ) -> Union[PhiTensor, GammaTensor]:
        # Last image that has been forward passed through this activation function
        self.last_forward = input_array

        gt = input_array > 0

        return input_array * gt + input_array * ((gt * -1) + 1) * self.slope

    def derivative(
        self, input_array: Optional[Union[PhiTensor, GammaTensor]] = None
    ) -> Union[PhiTensor, GammaTensor]:
        last_forward = input_array if input_array else self.last_forward

        if last_forward is None:
            raise ValueError(
                "last forward layer is None. Please provide an input to the derivative method."
            )

        res = (last_forward > 0).child * 1 + (last_forward <= 0).child * self.slope

        if isinstance(input_array, PhiTensor):
            return PhiTensor(
                child=res,
                data_subjects=last_forward.data_subjects,
                min_vals=last_forward.min_vals * 0,
                max_vals=last_forward.max_vals * 1,
            )
        elif isinstance(input_array, GammaTensor):
            return GammaTensor(
                child=res,
                data_subjects=last_forward.data_subjects,
                min_vals=last_forward.min_vals * 0,
                max_vals=last_forward.max_vals * 1,
            )
        else:
            raise NotImplementedError(
                f"Undefined behavior for type {type(input_array)}"
            )


def get(activation_name: Optional[str]) -> Optional[Activation]:

    activation_name_map: Dict[str, Type[Activation]] = {
        "leaky_relu": leaky_ReLU,
    }

    if activation_name is None:
        return None
    elif isinstance(activation_name, str):
        if activation_name not in activation_name_map:
            raise NotImplementedError(
                f"Activation {activation_name} is not current supported in PySyft."
            )
        else:
            activation_func = activation_name_map[activation_name]
            return activation_func()
    else:
        raise TypeError(
            f"Could not interpret activation function identifier: {activation_name}"
        )



# future
from __future__ import annotations

# stdlib
from typing import Any

# third party
from nacl.signing import VerifyKey


class AbstractDataSubjectLedger:
    store: AbstractLedgerStore
    user_key: Any

    def bind_to_store_with_key(
        self, store: AbstractLedgerStore, user_key: VerifyKey
    ) -> None:
        raise NotImplementedError


class AbstractLedgerStore:
    def get(self, key: VerifyKey) -> AbstractDataSubjectLedger:
        raise NotImplementedError

    def set(self, key: VerifyKey, value: AbstractDataSubjectLedger) -> None:
        raise NotImplementedError



# future
from __future__ import annotations

# stdlib
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
import numpy as np

# relative
from ...lib.numpy.array import capnp_deserialize
from ...lib.numpy.array import capnp_serialize
from ..common.serde.capnp import CapnpModule
from ..common.serde.capnp import chunk_bytes
from ..common.serde.capnp import combine_bytes
from ..common.serde.capnp import get_capnp_schema
from ..common.serde.capnp import serde_magic_header
from ..common.serde.deserialize import _deserialize as deserialize
from ..common.serde.serializable import serializable
from ..common.serde.serialize import _serialize as serialize
from .config import DEFAULT_FLOAT_NUMPY_TYPE
from .config import DEFAULT_INT_NUMPY_TYPE
from .passthrough import PassthroughTensor  # type: ignore
from .passthrough import is_acceptable_simple_type  # type: ignore
from .smpc import context


@serializable(capnp_bytes=True)
class FixedPrecisionTensor(PassthroughTensor):

    # __attr_allowlist__ = ("child", "_base", "_precision", "_scale")

    def __init__(
        self,
        value: Union[int, float, np.ndarray] = None,
        base: int = 2,
        precision: int = 16,
    ) -> None:
        self._base = base
        self._precision = precision
        self._scale = base**precision
        if value is not None:
            # TODO :Should modify to be compatiable with torch.

            super().__init__(self.encode(value))
        else:
            super().__init__(None)

    def encode(self, value: Union[int, float, np.ndarray]) -> np.ndarray:
        encoded_value = np.array(self._scale * value, DEFAULT_INT_NUMPY_TYPE)
        return encoded_value

    @property
    def precision(self) -> int:
        """Get the precision for the FixedPrecisionTensor.

        Returns:
            int: precision.
        """
        return self._precision

    @property
    def base(self) -> int:
        """Get the base for the FixedPrecisionTensor.

        Returns:
            int: base
        """
        return self._base

    @property
    def scale(self) -> int:
        """Get the scale for the FixedPrecisionTensor.

        Returns:
            int: the scale.
        """
        return self._scale

    @property
    def dtype(self) -> np.dtype:
        return getattr(self.child, "dtype", None)

    @property
    def shape(self) -> Tuple[Any, ...]:
        if self.child is None:
            return ()
        else:
            return self.child.shape

    def decode(self) -> Any:
        # relative
        from .smpc.share_tensor import ShareTensor

        value = self.child.child if isinstance(self.child, ShareTensor) else self.child

        correction = (value < 0).astype(DEFAULT_INT_NUMPY_TYPE)

        dividend = np.trunc(value / self._scale - correction)
        remainder = value % self._scale
        remainder += (
            (remainder == 0).astype(DEFAULT_INT_NUMPY_TYPE) * self._scale * correction
        )
        value = (
            dividend.astype(DEFAULT_FLOAT_NUMPY_TYPE)
            + remainder.astype(DEFAULT_FLOAT_NUMPY_TYPE) / self._scale
        )
        return value

    def sanity_check(
        self, other: Union[FixedPrecisionTensor, int, float, np.ndarray]
    ) -> FixedPrecisionTensor:
        if isinstance(other, FixedPrecisionTensor):
            if self.base != other.base or self.precision != other.precision:
                raise ValueError(
                    f"Base:{self.base,other.base} and Precision: "
                    + f"{self.precision, other.precision} should be same for "
                    + "computation on FixedPrecisionTensor"
                )
        elif is_acceptable_simple_type(other):
            other = FixedPrecisionTensor(
                value=other, base=self.base, precision=self.precision
            )
        else:
            raise ValueError(f"Invalid type for FixedPrecisionTensor: {type(other)}")

        return other

    def __add__(self, other: Any) -> FixedPrecisionTensor:
        res = FixedPrecisionTensor(base=self._base, precision=self._precision)
        if isinstance(other, np.ndarray) and other.dtype == np.dtype("bool"):
            res.child = self.child + other
        else:
            other = self.sanity_check(other)
            res.child = self.child + other.child
        return res

    def __sub__(self, other: Any) -> FixedPrecisionTensor:
        res = FixedPrecisionTensor(base=self._base, precision=self._precision)
        if isinstance(other, np.ndarray) and other.dtype == np.dtype("bool"):
            res.child = self.child - other
        else:
            other = self.sanity_check(other)
            res.child = self.child - other.child
        return res

    def __mul__(self, other: Any) -> FixedPrecisionTensor:
        res = FixedPrecisionTensor(base=self._base, precision=self._precision)
        if isinstance(other, np.ndarray) and other.dtype == np.dtype("bool"):
            res.child = self.child * other
        else:
            other = self.sanity_check(other)
            context.FPT_CONTEXT["seed_id_locations"] = context.SMPC_CONTEXT.get(
                "seed_id_locations", None
            )
            res.child = self.child * other.child
            res = res / self.scale

        return res

    def __matmul__(self, other: Any) -> FixedPrecisionTensor:
        res = FixedPrecisionTensor(base=self._base, precision=self._precision)
        if isinstance(other, np.ndarray) and other.dtype == np.dtype("bool"):
            raise ValueError("Should not get  a boolan array to matmul")
        else:
            other = self.sanity_check(other)
            context.FPT_CONTEXT["seed_id_locations"] = context.SMPC_CONTEXT.get(
                "seed_id_locations", None
            )
            res.child = self.child @ other.child
            res = res / self.scale

        return res

    def __rmatmul__(self, other: Any) -> FixedPrecisionTensor:
        res = FixedPrecisionTensor(base=self._base, precision=self._precision)
        if isinstance(other, np.ndarray) and other.dtype == np.dtype("bool"):
            raise ValueError("Should not get  a boolan array to matmul")
        else:
            other = self.sanity_check(other)
            context.FPT_CONTEXT["seed_id_locations"] = context.SMPC_CONTEXT.get(
                "seed_id_locations", None
            )
            res.child = self.child.__rmatmul__(other.child)
            res = res / self.scale

        return res

    def __truediv__(
        self, other: Union[int, np.integer, FixedPrecisionTensor]
    ) -> FixedPrecisionTensor:
        if isinstance(other, FixedPrecisionTensor):
            raise ValueError("We do not support Private Division yet.")

        res = FixedPrecisionTensor(base=self._base, precision=self._precision)
        if isinstance(self.child, np.ndarray) or np.isscalar(self.child):
            res.child = np.trunc(self.child / other).astype(DEFAULT_INT_NUMPY_TYPE)
        else:
            res.child = self.child / other
        return res

    def transpose(
        self, *args: List[Any], **kwargs: Dict[Any, Any]
    ) -> FixedPrecisionTensor:
        res = FixedPrecisionTensor(base=self._base, precision=self._precision)
        res.child = self.child.transpose(*args, **kwargs)
        return res

    @property
    def T(self) -> FixedPrecisionTensor:
        return self.transpose()

    # For comparison , we could skip conversion to FPT , as operation holds without encoding.
    def __lt__(self, other: Any) -> FixedPrecisionTensor:
        other = self.sanity_check(other)
        value = (self.child < other.child) * 1

        res = FixedPrecisionTensor(
            value=value, base=self._base, precision=self._precision
        )
        return res

    def __gt__(self, other: Any) -> FixedPrecisionTensor:
        other = self.sanity_check(other)
        value = (self.child > other.child) * 1
        res = FixedPrecisionTensor(
            value=value, base=self._base, precision=self._precision
        )
        return res

    def concatenate(
        self, other: FixedPrecisionTensor, *args: List[Any], **kwargs: Dict[str, Any]
    ) -> FixedPrecisionTensor:
        if not isinstance(other, FixedPrecisionTensor):
            raise NotImplementedError

        res = FixedPrecisionTensor(base=self._base, precision=self._precision)
        res.child = self.child.concatenate(other.child, *args, **kwargs)

        return res

    def all(self) -> bool:
        return self.child.all()

    def copy(self, order: Optional[str] = "K") -> FixedPrecisionTensor:
        res = FixedPrecisionTensor(base=self._base, precision=self._precision)
        res.child = self.child.copy(order=order)
        return res

    def sum(
        self, axis: Optional[Union[int, Tuple[int, ...]]] = None
    ) -> FixedPrecisionTensor:
        res = FixedPrecisionTensor(base=self._base, precision=self._precision)
        if isinstance(self.child, np.ndarray):
            res.child = np.array(self.child.sum(axis=axis))
        else:
            res.child = self.child.sum(axis=axis)
        return res

    def _object2bytes(self) -> bytes:
        schema = get_capnp_schema(schema_file="fixed_precision_tensor.capnp")

        fpt_struct: CapnpModule = schema.FPT  # type: ignore
        fpt_msg = fpt_struct.new_message()
        # this is how we dispatch correct deserialization of bytes
        fpt_msg.magicHeader = serde_magic_header(type(self))

        # child of FPT tensor could either be ShareTensor or np.ndarray
        if isinstance(self.child, np.ndarray) or np.isscalar(self.child):
            chunk_bytes(
                capnp_serialize(np.array(self.child), to_bytes=True), "child", fpt_msg
            )
            fpt_msg.isNumpy = True
        else:
            chunk_bytes(serialize(self.child, to_bytes=True), "child", fpt_msg)  # type: ignore
            fpt_msg.isNumpy = False

        fpt_msg.base = self.base
        fpt_msg.precision = self.precision

        # to pack or not to pack?
        # to_bytes = fpt_msg.to_bytes()

        return fpt_msg.to_bytes_packed()

    @staticmethod
    def _bytes2object(buf: bytes) -> FixedPrecisionTensor:
        schema = get_capnp_schema(schema_file="fixed_precision_tensor.capnp")
        fpt_struct: CapnpModule = schema.FPT  # type: ignore
        # https://stackoverflow.com/questions/48458839/capnproto-maximum-filesize
        MAX_TRAVERSAL_LIMIT = 2**64 - 1
        # to pack or not to pack?
        # fpt_msg = fpt_struct.from_bytes(buf, traversal_limit_in_words=2 ** 64 - 1)
        fpt_msg = fpt_struct.from_bytes_packed(
            buf, traversal_limit_in_words=MAX_TRAVERSAL_LIMIT
        )

        if fpt_msg.isNumpy:
            child = capnp_deserialize(combine_bytes(fpt_msg.child), from_bytes=True)
        else:
            child = deserialize(combine_bytes(fpt_msg.child), from_bytes=True)

        base = fpt_msg.base
        precision = fpt_msg.precision

        res = FixedPrecisionTensor(base=base, precision=precision)
        res.child = child
        return res

    def __getitem__(
        self, item: Union[str, int, slice, PassthroughTensor]
    ) -> FixedPrecisionTensor:
        res = FixedPrecisionTensor(base=self.base, precision=self.precision)
        if isinstance(item, PassthroughTensor):
            res.child = self.child.getitem(item.child)
        else:
            res.child = self.child.getitem(item)

        return res

    def one_hot(self) -> FixedPrecisionTensor:
        if not isinstance(self.child, np.ndarray):
            raise ValueError("One hot work only for numpy array child type")

        value = self.decode().astype(DEFAULT_INT_NUMPY_TYPE)
        one_hot_Y = np.zeros((value.size, value.max() + 1))
        one_hot_Y[np.arange(value.size), value] = 1
        one_hot_Y = one_hot_Y.T

        res = FixedPrecisionTensor(
            value=one_hot_Y, base=self.base, precision=self.precision
        )
        return res



# future
from __future__ import annotations

# stdlib
import time
from typing import Any
from typing import Dict
from typing import Optional

# third party
from nacl.signing import VerifyKey
from typing_extensions import final

# relative
from .....common.serde.serializable import serializable
from ....abstract.node import AbstractNode
from ..generic_payload.messages import GenericPayloadMessage
from ..generic_payload.messages import GenericPayloadMessageWithReply
from ..generic_payload.messages import GenericPayloadReplyMessage


@serializable(recursive_serde=True)
@final
class SleepMessage(GenericPayloadMessage):
    ...


@serializable(recursive_serde=True)
@final
class SleepReplyMessage(GenericPayloadReplyMessage):
    ...


@serializable(recursive_serde=True)
@final
class SleepMessageWithReply(GenericPayloadMessageWithReply):
    message_type = SleepMessage
    message_reply_type = SleepReplyMessage

    def run(
        self, node: AbstractNode, verify_key: Optional[VerifyKey] = None
    ) -> Dict[str, Any]:
        seconds = float(self.kwargs["seconds"])
        time.sleep(seconds)
        return {
            "seconds": seconds,
            "result": f"Finished Sleeping for {seconds}",
            "status_code": 200,
        }



# stdlib
from typing import Any
from typing import Dict
from typing import Union

# third party
from packaging import version
import torch

# relative
from . import device  # noqa: 401
from . import parameter  # noqa: 401
from . import return_types  # noqa: 401
from . import size  # noqa: 401
from . import uppercase_tensor  # noqa: 401
from ...ast import add_dynamic_objects
from ...ast.globals import Globals
from ...logger import info
from .allowlist import allowlist
from .allowlist import dynamic_allowlist

TORCH_VERSION = version.parse(torch.__version__.split("+")[0])


def get_return_type(support_dict: Union[str, Dict[str, str]]) -> str:
    if isinstance(support_dict, str):
        return support_dict
    else:
        return support_dict["return_type"]


def version_supported(support_dict: Union[str, Dict[str, str]]) -> bool:
    if isinstance(support_dict, str):
        return True
    else:
        # if we are on either side of the min or max versions we don't support this op
        if "min_version" in support_dict and TORCH_VERSION < version.parse(
            support_dict["min_version"]
        ):
            return False
        if "max_version" in support_dict and TORCH_VERSION > version.parse(
            support_dict["max_version"]
        ):
            return False
        return True


def create_torch_ast(client: Any = None) -> Globals:
    ast = Globals(client)

    # most methods work in all versions and have a single return type
    # for the more complicated ones we pass a dict with keys like return_type and
    # min_version
    for method, return_type_name_or_dict in allowlist.items():
        if version_supported(support_dict=return_type_name_or_dict):
            return_type = get_return_type(support_dict=return_type_name_or_dict)
            if return_type == "unknown":
                # this allows us to import them for testing
                continue
            ast.add_path(
                path=method, framework_reference=torch, return_type_name=return_type
            )
            # add all the torch.nn.Parameter hooks
            if method.startswith("torch.Tensor."):
                method = method.replace("torch.Tensor.", "torch.nn.Parameter.")
                return_type = return_type.replace("torch.Tensor", "torch.nn.Parameter")
                ast.add_path(
                    path=method, framework_reference=torch, return_type_name=return_type
                )
        else:
            info(f"Skipping {method} not supported in {TORCH_VERSION}")

    add_dynamic_objects(ast, list(dynamic_allowlist.items()))

    for klass in ast.classes:
        klass.create_pointer_class()
        klass.create_send_method()
        klass.create_storable_object_attr_convenience_methods()

    return ast



# stdlib
from typing import List
from typing import Optional
from typing import Type

# third party
from nacl.signing import VerifyKey

# relative
from ......util import traceback_and_raise
from .....common.group import VERIFYALL
from ....abstract.node import AbstractNode
from ..auth import service_auth
from ..node_service import ImmediateNodeServiceWithReply
from .get_repr_messages import GetReprMessage
from .get_repr_messages import GetReprReplyMessage


class GetReprService(ImmediateNodeServiceWithReply):
    @staticmethod
    @service_auth(root_only=True)
    def process(
        node: AbstractNode,
        msg: GetReprMessage,
        verify_key: Optional[VerifyKey] = None,
    ) -> GetReprReplyMessage:
        if verify_key is None:
            traceback_and_raise(
                "Can't process an GetReprService with no verification key."
            )

        obj = node.store.get(msg.id_at_location, proxy_only=True)
        contains_all_in_permissions = any(
            key is VERIFYALL for key in obj.read_permissions.keys()
        )

        if not (
            verify_key in obj.read_permissions.keys()
            or verify_key == node.root_verify_key
            or contains_all_in_permissions
        ):
            raise PermissionError("Permission to get repr of object not granted!")
        else:
            # TODO: Create a remote print interface for objects which displays them in a
            # nice way, we could also even buffer this between chained ops until we
            # return so that we can print once and display a nice list of data and ops
            # issue: https://github.com/OpenMined/PySyft/issues/5167
            result = repr(obj.data)
            return GetReprReplyMessage(repr=result, address=msg.reply_to)

    @staticmethod
    def message_handler_types() -> List[Type[GetReprMessage]]:
        return [GetReprMessage]



# stdlib
from enum import Enum
import inspect
from typing import Any
from typing import Dict
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
from google.protobuf.reflection import GeneratedProtocolMessageType
from nacl.signing import VerifyKey

# syft absolute
import syft as sy

# relative
from ..... import lib
from .....proto.core.node.common.action.get_set_property_pb2 import (
    GetOrSetPropertyAction as GetOrSetPropertyAction_PB,
)
from .....util import inherit_tags
from ....common.serde.serializable import serializable
from ....common.uid import UID
from ....io.address import Address
from ....store.storeable_object import StorableObject
from ...abstract.node import AbstractNode
from .common import ImmediateActionWithoutReply
from .greenlets_switch import retrieve_object
from .run_class_method_action import RunClassMethodAction


class PropertyActions(Enum):
    SET = 1
    GET = 2
    DEL = 3


@serializable()
class GetOrSetPropertyAction(ImmediateActionWithoutReply):
    def __init__(
        self,
        path: str,
        _self: Any,
        id_at_location: UID,
        address: Address,
        args: Union[Tuple[Any, ...], List[Any]],
        kwargs: Dict[Any, Any],
        action: PropertyActions,
        map_to_dyn: bool,
        set_arg: Optional[Any] = None,
        msg_id: Optional[UID] = None,
    ):
        super().__init__(address, msg_id=msg_id)
        self.path = path
        self.id_at_location = id_at_location
        self.set_arg = set_arg
        self._self = _self
        self.action = action
        self.args = args
        self.kwargs = kwargs
        self.map_to_dyn = map_to_dyn

    def intersect_keys(
        self,
        left: Dict[VerifyKey, Optional[UID]],
        right: Dict[VerifyKey, Optional[UID]],
    ) -> Dict[VerifyKey, Optional[UID]]:
        return RunClassMethodAction.intersect_keys(left, right)

    def execute_action(self, node: AbstractNode, verify_key: VerifyKey) -> None:
        ast_node = node.lib_ast.query(self.path)
        method = ast_node.object_ref

        # storable object raw from object store
        resolved_self = retrieve_object(node, self._self.id_at_location, self.path)
        result_read_permissions = resolved_self.read_permissions

        resolved_args = []
        tag_args = []
        for arg in self.args:
            r_arg = retrieve_object(node, arg.id_at_location, self.path)
            result_read_permissions = self.intersect_keys(
                result_read_permissions, r_arg.read_permissions
            )
            tag_args.append(r_arg)
            resolved_args.append(r_arg.data)

        resolved_kwargs = {}
        tag_kwargs = {}
        for arg_name, arg in self.kwargs.items():
            r_arg = retrieve_object(node, arg.id_at_location, self.path)
            result_read_permissions = self.intersect_keys(
                result_read_permissions, r_arg.read_permissions
            )
            tag_kwargs[arg_name] = r_arg
            resolved_kwargs[arg_name] = r_arg.data

        if not (inspect.isdatadescriptor(method) or self.map_to_dyn):
            raise ValueError(f"{method} not an actual property!")

        (
            upcasted_args,
            upcasted_kwargs,
        ) = lib.python.util.upcast_args_and_kwargs(resolved_args, resolved_kwargs)

        data = resolved_self.data

        if self.map_to_dyn:
            if self.action == PropertyActions.SET:
                setattr(data, ast_node.name, *upcasted_args)
                result = None
                # since we may have changed resolve_self.data we need to check it back in
                node.store[self._self.id_at_location] = resolved_self
            elif self.action == PropertyActions.GET:
                result = getattr(data, ast_node.name)
            elif self.action == PropertyActions.DEL:
                raise ValueError(f"{self.action} not a valid action!")
        else:
            if self.action == PropertyActions.SET:
                result = method.__set__(data, *upcasted_args, **upcasted_kwargs)

                # since we may have changed resolve_self.data we need to check it back in
                node.store[self._self.id_at_location] = resolved_self

            elif self.action == PropertyActions.GET:
                result = method.__get__(data, *upcasted_args, **upcasted_kwargs)
            elif self.action == PropertyActions.DEL:
                result = method.__del__(data, *upcasted_args, **upcasted_kwargs)

                # since we may have changed resolve_self.data we need to check it back in
                node.store[self._self.id_at_location] = resolved_self
            else:
                raise ValueError(f"{self.action} not a valid action!")

        if lib.python.primitive_factory.isprimitive(value=result):
            result = lib.python.primitive_factory.PrimitiveFactory.generate_primitive(
                value=result, id=self.id_at_location
            )
        else:
            if hasattr(result, "id"):
                # just to make mypy to shut up, impossible case
                if result is None:
                    raise Exception("Please convert None so _SyNone.")

                try:
                    if hasattr(result, "_id"):
                        # set the underlying id
                        result._id = self.id_at_location
                    else:
                        result.id = self.id_at_location

                    if result.id != self.id_at_location:
                        raise AttributeError("IDs don't match")
                except AttributeError:
                    raise Exception("MAKE VALID SCHEMA")

        if not isinstance(result, StorableObject):
            result = StorableObject(
                id=self.id_at_location,
                data=result,
                read_permissions=result_read_permissions,
            )

        # When GET, result is a new object, we give new tags to it
        if self.action == PropertyActions.GET:
            inherit_tags(
                attr_path_and_name=self.path,
                result=result,
                self_obj=resolved_self,
                args=tag_args,
                kwargs=tag_kwargs,
            )

        node.store[self.id_at_location] = result

    def __repr__(self) -> str:
        attr_name = self.path.split(".")[-1]
        self_name = str(self._self.__class__.__name__)

        if self.action == PropertyActions.SET:
            val = self.args[0]
            return f"GetOrSetPropertyAction {self_name}.{attr_name} = {val}"
        elif self.action == PropertyActions.GET:
            return f"GetOrSetPropertyAction GET {self_name}.{attr_name}"
        else:
            return f"GetOrSetPropertyAction DEL {self_name}.{attr_name}"

    def _object2proto(self) -> GetOrSetPropertyAction_PB:
        """Returns a protobuf serialization of self.
        As a requirement of all objects which inherit from Serializable,
        this method transforms the current object into the corresponding
        Protobuf object so that it can be further serialized.
        :return: returns a protobuf object
        :rtype: GetOrSetPropertyAction_PB
        .. note::
            This method is purely an internal method. Please use sy.serialize(object) or one of
            the other public serialization methods if you wish to serialize an
            object.
        """
        return GetOrSetPropertyAction_PB(
            path=self.path,
            id_at_location=sy.serialize(self.id_at_location),
            args=list(map(lambda x: sy.serialize(x), self.args)),
            kwargs={k: sy.serialize(v) for k, v in self.kwargs.items()},
            address=sy.serialize(self.address),
            _self=sy.serialize(self._self, to_bytes=True),
            msg_id=sy.serialize(self.id),
            action=self.action.value,
            map_to_dyn=self.map_to_dyn,
        )

    @staticmethod
    def _proto2object(
        proto: GetOrSetPropertyAction_PB,
    ) -> "GetOrSetPropertyAction":
        """Creates a GetOrSetPropertyAction from a protobuf
        As a requirement of all objects which inherit from Serializable,
        this method transforms a protobuf object into an instance of this class.
        :return: returns an instance of GetOrSetPropertyAction
        :rtype: GetOrSetPropertyAction
        .. note::
            This method is purely an internal method. Please use syft.deserialize()
            if you wish to deserialize an object.
        """

        return GetOrSetPropertyAction(
            path=proto.path,
            id_at_location=sy.deserialize(blob=proto.id_at_location),
            address=sy.deserialize(blob=proto.address),
            _self=sy.deserialize(blob=proto._self, from_bytes=True),
            msg_id=sy.deserialize(blob=proto.msg_id),
            args=tuple(sy.deserialize(blob=x) for x in proto.args),
            kwargs={k: sy.deserialize(blob=v) for k, v in proto.kwargs.items()},
            action=PropertyActions(proto.action),
            map_to_dyn=proto.map_to_dyn,
        )

    @staticmethod
    def get_protobuf_schema() -> GeneratedProtocolMessageType:
        """Return the type of protobuf object which stores a class of this type
        As a part of serialization and deserialization, we need the ability to
        lookup the protobuf object type directly from the object type. This
        static method allows us to do this.
        Importantly, this method is also used to create the reverse lookup ability within
        the metaclass of Serializable. In the metaclass, it calls this method and then
        it takes whatever type is returned from this method and adds an attribute to it
        with the type of this class attached to it. See the MetaSerializable class for details.
        :return: the type of protobuf object which corresponds to this class.
        :rtype: GeneratedProtocolMessageType
        """

        return GetOrSetPropertyAction_PB



# stdlib
from typing import Dict as TypeDict
from typing import List as TypeList
from typing import Optional
from typing import Type

# third party
from nacl.signing import VerifyKey

# relative
from ......logger import traceback_and_raise  # type: ignore
from .....adp.data_subject_ledger import DataSubjectLedger  # type: ignore
from .....common.uid import UID  # type: ignore
from .....store.storeable_object import StorableObject  # type: ignore
from ....abstract.node import AbstractNode  # type: ignore
from ...action.greenlets_switch import retrieve_object  # type: ignore

# from ...action import context  # type: ignore
from ..node_service import ImmediateNodeServiceWithoutReply  # type: ignore
from .publish_messages import PublishScalarsAction  # type: ignore


class PublishScalarsService(ImmediateNodeServiceWithoutReply):
    @staticmethod
    def process(
        node: AbstractNode, msg: PublishScalarsAction, verify_key: VerifyKey
    ) -> None:
        # get scalar objects from store
        results = (
            []
        )  # TODO: Ask Andrew why this is a thing- are we able to publish multiple things at once?
        for publish_id in msg.publish_ids_at_location:
            # print("PublishScalarsService:33")
            print("Publish_id: ", publish_id)
            try:
                # print("PublishScalarsService:36: TRY: publish_object = node.store.get(publish_id)")
                # in memory cache for faster fetch of large objects
                # if publish_id.no_dash in context.OBJ_CACHE:
                #     publish_object = context.OBJ_CACHE[publish_id.no_dash]
                # else:
                publish_object = retrieve_object(
                    node=node,
                    id_at_location=publish_id,
                    path="Publish object path",
                    proxy_only=False,
                )
                # publish_object = node.store.get(publish_id)
                if hasattr(publish_object, "data"):
                    publish_object = publish_object.data
                try:
                    ledger = DataSubjectLedger.get_or_create(
                        store=node.ledger_store, user_key=verify_key
                    )
                    if ledger is None:
                        raise Exception("Unable to get ledger so we cannot publish")
                except Exception as e:
                    print(f"Failed to get a ledger. {e}")
                    raise e

                result = publish_object.publish(
                    deduct_epsilon_for_user=node.users.deduct_epsilon_for_user,
                    get_budget_for_user=node.users.get_budget_for_user,
                    ledger=ledger,
                    sigma=msg.sigma,
                )

                # print("PublishScalarsService:44: SUCCESS: publish_object.publish()")
                results.append(result)
            except Exception as e:
                # print("PublishScalarsService:53: EXCEPTION - missing id")
                log = (
                    f"Unable to Get Object with ID {publish_id} from store. "
                    + f"Possible dangling Pointer. {e}"
                )
                traceback_and_raise(Exception(log))

        # give the caller permission to download this
        read_permissions: TypeDict[VerifyKey, Optional[UID]] = {verify_key: None}
        search_permissions: TypeDict[VerifyKey, Optional[UID]] = {verify_key: None}

        if len(results) == 1:
            results = results[0]

        storable = StorableObject(
            id=msg.id_at_location,
            data=results,
            description=f"Approved AutoDP Result: {msg.id_at_location}",
            read_permissions=read_permissions,
            search_permissions=search_permissions,
        )
        node.store[msg.id_at_location] = storable

    @staticmethod
    def message_handler_types() -> TypeList[Type[PublishScalarsAction]]:
        return [PublishScalarsAction]



# stdlib
from typing import List
from typing import Set
from typing import Type

# third party
from nacl.signing import VerifyKey

# relative
from ......logger import error
from .....common.message import ImmediateSyftMessageWithReply
from .....common.uid import UID
from ....abstract.node import AbstractNode
from ....domain_client import DomainClient
from ..auth import service_auth
from ..node_service import ImmediateNodeServiceWithReply
from ..peer_discovery.peer_discovery_messages import node_id_to_peer_route_metadata
from .network_search_messages import NetworkSearchMessage
from .network_search_messages import NetworkSearchResponse


class NetworkSearchService(ImmediateNodeServiceWithReply):
    @staticmethod
    @service_auth(guests_welcome=True)
    def process(
        node: AbstractNode,
        msg: NetworkSearchMessage,
        verify_key: VerifyKey,
    ) -> NetworkSearchResponse:
        queries = set(msg.content)

        # refresh any missing peer clients
        node.reload_peer_clients()  # type: ignore

        def query_client_store(client: DomainClient, tags: Set[str]) -> bool:
            # source is the domain in an association
            for data in client.store:
                if tags.issubset(set(data.tags)):
                    return True
            return False

        tested_nodes: Set[UID] = set()
        matching_nodes: Set[UID] = set()
        for node_id, clients in node.all_peer_clients().items():  # type: ignore
            # try each client / route one after the next
            for client in clients:
                try:
                    # only check a client once
                    if node_id not in tested_nodes and query_client_store(
                        client=client, tags=queries
                    ):
                        matching_nodes.add(node_id)
                except Exception as e:
                    error(f"Failed to query {node_id} with {client}. {e}")

        peer_routes = []
        for node_id in matching_nodes:
            node_row = node.node.first(node_uid=node_id.no_dash)  # type: ignore
            if node_row:
                peer_routes += node_id_to_peer_route_metadata(
                    node=node, node_row=node_row  # type: ignore
                )

        return NetworkSearchResponse(
            address=msg.reply_to,
            status_code=200,
            content={"status": "ok", "data": peer_routes},
        )

    @staticmethod
    def message_handler_types() -> List[Type[ImmediateSyftMessageWithReply]]:
        return [NetworkSearchMessage]
